package gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import util.*;
import simulator.Simulation;
import simulator.SimulationParameters;

abstract class SimulationWindow extends JFrame implements WindowListener {
	private static String VERSION = "Cellsim 1.8";
	private float xangle = 90.0f, yangle = 45.0f, zangle = 30.0f;
	private JButton buttonAdvance = new JButton("Advance");
	private JButton buttonKill = new JButton("Kill");
	private JButton buttonRefresh = new JButton("Refresh");

	double panx = 0.0;
	double pany = 0.0;
	double panz = 0.0;
	double pandelta = 0.1;

	double rotx = 0.0;
	double roty = 0.0;
	double rotz = 0.0;
	double rotdelta = 0.2;

	private JSlider panxSlider =
		new JSlider(JSlider.HORIZONTAL, -5, 5, (int) panx);
	private JSlider panySlider =
		new JSlider(JSlider.HORIZONTAL, -5, 5, (int) pany);
	private JSlider panzSlider =
		new JSlider(JSlider.HORIZONTAL, -5, 5, (int) panz);

	private JSlider rotxSlider =
		new JSlider(JSlider.HORIZONTAL, -180, 180, (int) rotx);
	private JSlider rotySlider =
		new JSlider(JSlider.HORIZONTAL, -180, 180, (int) roty);
	private JSlider rotzSlider =
		new JSlider(JSlider.HORIZONTAL, -180, 180, (int) rotz);

	JButton buttonRotateX = new JButton("X");
	JButton buttonRotateY = new JButton("Y");
	JButton buttonRotateZ = new JButton("Z");

	protected JButton buttonStep = new JButton("Step");
	protected Simulation sim;
	private StatsTableModel statsTable;
	private Integer windowID;
	private boolean autoscale = false;

	class SliderListener implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();

			if (!source.getValueIsAdjusting()) {
				int fps = (int) source.getValue();

				if (source == panxSlider) {
					panx = (float) fps;
				} else if (source == panySlider) {
					pany = (float) fps;
				} else if (source == panzSlider) {
					panz = (float) fps;
				} else if (source == rotxSlider) {
					rotx = (float) fps;
				} else if (source == rotySlider) {
					rotx = 0;
					roty = (float) fps;
				} else if (source == rotzSlider) {
					roty = 0;
					rotx = 0;
					rotz = (float) fps;
				} else {
					Assert.badBranch("Bad slider!");
				}

				refreshState();
			}
		}
	}

	class ButtonItemListener implements ActionListener {
		public void actionPerformed(ActionEvent e) {
			JButton button = (JButton) e.getSource();

			if (button == buttonStep) {
				nextStep();
			} else if (button == buttonAdvance) {
				advanceSteps();
			} else if (button == buttonKill) {
				String input = null;
				input =
					JOptionPane.showInputDialog(
						getParent(),
						"Enter fraction of cells (between 0.0 and 1.0 to randomly kill:",
						"",
						JOptionPane.QUESTION_MESSAGE);
				if ((input != null) && (input.length() != 0)) {
					double fraction = 0.0;
					try {
						fraction = (new Double(input)).doubleValue();
					} catch (NumberFormatException nfe) {
						Report.nonFatalError("Enter a number: " + input);
						return;
					}

					if (fraction >= 0.0 && fraction <= 1.0) {
						String state = null;
						state =
							(String) JOptionPane.showInputDialog(
								getParent(),
								"Enter state to randomly kill (or null for all):",
								"",
								JOptionPane.QUESTION_MESSAGE);

						if (state.length() == 0) {
							state = null;
						}

						System.err.println("Killing randomly: " + fraction);
						sim.getSystemModel().killRandomly(fraction, state);
						refreshState();
					} else {
						System.err.println("Bad input: " + fraction);
					}
				}
			} else if (button == buttonRefresh) {
				refreshState();
			} else {
				Assert.badBranch("Unrecognized button!");
			}
		}
	}

	SimulationWindow(SimulationParameters params) {
		super(VERSION);		
		JMenuBar mainMenu = new MainMenu(this);
		setJMenuBar(mainMenu);

		sim = new Simulation(params);
		setTitle(sim.getTitle());

		if (params.getAutoScale()) {
			panz = 0;
		}

		setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
		addWindowListener(this);

		/* Create split windows for 3D display, statistics, output and error messages*/
		Container mainContainer = getContentPane();
		mainContainer.setLayout(new BorderLayout(1, 1));	
		
		//JPanel pane = new JPanel();
		//pane.setLayout(new BorderLayout());
		//mainContainer.add(pane, BorderLayout.WEST);

		if (params.isCompact()) {

		} else {
			JPanel controlPanel = new JPanel();

			controlPanel.setLayout(
				new GridLayout(0, 1, 0, 2));

			JToolBar toolbar = new JToolBar();

			toolbar.add(buttonStep);
			toolbar.add(buttonAdvance);
			toolbar.add(buttonKill);
			toolbar.add(buttonRefresh);

			toolbar.add(buttonRotateX);
			toolbar.add(buttonRotateY);
			toolbar.add(buttonRotateZ);

			toolbar.setAlignmentX(Component.CENTER_ALIGNMENT);
			controlPanel.add(toolbar);
			
			JToolBar rotbar = new JToolBar();
			rotbar.setBorder(BorderFactory.createTitledBorder("Rotate"));
			rotbar.setLayout(new BoxLayout(rotbar, BoxLayout.Y_AXIS));
			rotxSlider.setBorder(BorderFactory.createTitledBorder("X"));
			rotxSlider.addChangeListener(new SliderListener());
			rotbar.add(rotxSlider);
			rotySlider.addChangeListener(new SliderListener());
			rotySlider.setBorder(BorderFactory.createTitledBorder("Y"));
			rotbar.add(rotySlider);
			rotzSlider.setBorder(BorderFactory.createTitledBorder("Z"));
			rotzSlider.setMajorTickSpacing(90);
			rotzSlider.setMinorTickSpacing(5);
			rotzSlider.setPaintTicks(true);
			rotzSlider.setPaintLabels(true);
			rotzSlider.addChangeListener(new SliderListener());
			rotbar.add(rotzSlider);

			JToolBar panbar = new JToolBar();
			panbar.setBorder(BorderFactory.createTitledBorder("Move"));
			panbar.setLayout(new BoxLayout(panbar, BoxLayout.Y_AXIS));
			panxSlider.addChangeListener(new SliderListener());
			panxSlider.setBorder(BorderFactory.createTitledBorder("X"));
			panxSlider.addChangeListener(new SliderListener());
			panbar.add(panxSlider);
			panySlider.addChangeListener(new SliderListener());
			panySlider.setBorder(BorderFactory.createTitledBorder("Y"));
			panySlider.addChangeListener(new SliderListener());
			panbar.add(panySlider);
			panzSlider.addChangeListener(new SliderListener());
			panzSlider.setBorder(BorderFactory.createTitledBorder("Z"));
			panzSlider.addChangeListener(new SliderListener());
			panzSlider.setMajorTickSpacing(10);
			panzSlider.setMinorTickSpacing(5);
			panzSlider.setPaintTicks(true);
			panzSlider.setPaintLabels(true);
			panbar.add(panzSlider);

			buttonStep.addActionListener(new ButtonItemListener());
			buttonAdvance.addActionListener(new ButtonItemListener());
			buttonKill.addActionListener(new ButtonItemListener());
			buttonRefresh.addActionListener(new ButtonItemListener());

			statsTable = new StatsTableModel();
			JTable stable = new JTable(statsTable);
			// stable.setMaximumSize (new Dimension (200, 200));
			JScrollPane p2 = new JScrollPane(stable);
			JScrollPane p3 = new JScrollPane(new JTextArea());
			JScrollPane p4 = new JScrollPane(new JTextArea());
			p2.setSize(new Dimension(100, 120));
			JTabbedPane jtp = new JTabbedPane();
			jtp.addTab("Statistics", p2);
			jtp.addTab("Output", p3);
			jtp.addTab("Errors", p4);
			p2.setAlignmentX(Component.CENTER_ALIGNMENT);
			
			rotbar.setAlignmentX(Component.LEFT_ALIGNMENT);
			controlPanel.add(rotbar);

			panbar.setAlignmentX(Component.LEFT_ALIGNMENT);
			controlPanel.add(panbar);
			controlPanel.add(stable);
			
			controlPanel.setSize(new Dimension(300, mainContainer.getHeight()));
			mainContainer.add(controlPanel, BorderLayout.LINE_START);
		}
	}

	public void advanceSteps() {
		String input = null;
		input =
			JOptionPane.showInputDialog(
				getParent(),
				"Advance how many steps?",
				"",
				JOptionPane.QUESTION_MESSAGE);
		if ((input != null) && (input.length() != 0)) {
			{
				gotoStep((new Integer(input)).intValue() + sim.getStep());
			}
		}
	}

	void gotoStep(int stepNumber) {
		while (sim.getStep() < stepNumber) {
			nextStep();
		}
	}

	//	public void internalFrameActivated(InternalFrameEvent e) {
	//		simWinMgr.setCurrentSimulationWindow(windowID);
	//	}
	//
	//	public void internalFrameClosed(InternalFrameEvent e) {}
	//
	//	/* Implementing the InternalFrameListener */
	//	public void internalFrameClosing(InternalFrameEvent e) {
	//		simWinMgr.deleteSimulationWindow(windowID);
	//	}

	public void internalFrameDeactivated(InternalFrameEvent e) {
	}

	public void internalFrameDeiconified(InternalFrameEvent e) {
	}

	public void internalFrameIconified(InternalFrameEvent e) {
	}

	public void internalFrameOpened(InternalFrameEvent e) {
	}

	public abstract void refreshDisplay ();
	
	public void refreshState() {
		if (statsTable != null) {
			statsTable.updateStatistics(sim.getStatistics());
		}
		setTitle(sim.getTitle());
		refreshDisplay ();
	}

	public void setTitle(String s) {
		super.setTitle(VERSION + " - " + s);
	}
	void nextStep() {
		sim.nextStep();
		refreshState();
	}

	public void windowClosing(WindowEvent e) {
		dispose();
		System.exit(0);
	}

	public void windowOpened(WindowEvent e) {
	}
	public void windowIconified(WindowEvent e) {
	}
	public void windowDeiconified(WindowEvent e) {
	}
	public void windowActivated(WindowEvent e) {
	}
	public void windowDeactivated(WindowEvent e) {
	}
	public void windowClosed(WindowEvent e) {
	}

}
