package parsers.CPA;

import java.util.*;
import util.*;
import simulator.*;

public class CPACellProgram extends CellProgram {
	private boolean resolved;
	private Hashtable stateTable;
	private Hashtable chemicalTable;
	private Vector firstStates;

	public CPACellProgram() {
		stateTable = new Hashtable();
		chemicalTable = new Hashtable();
		firstStates = new Vector();
		resolved = false;
	}

	private void Report(String msg) {
		// System.err.println ("[CPA] " + msg);
		// System.err.flush ();
	}

	public CellState getState(String state) {
		CellState res = (CellState) stateTable.get(state);
		if (res == null) {
			Report.fatalError("Unrecognized state: " + state);
		}

		return res;
	}

	public CellState[] getStates(String[] states) {
		CellState[] cellStates = new CellState[states.length];

		for (int pos = 0; pos < states.length; pos++) {
			cellStates[pos] = getState(states[pos]);
		}
		return cellStates;
	}

	public ChemicalDescription getChemical(String chem) {
		ChemicalDescription res = (ChemicalDescription) chemicalTable.get(chem);
		return res; // Okay to return null for missing chemical	
	}

	public void extend(CellProgram cpa)
	// Combine all states of cpa into this
	{
		if (cpa instanceof CPACellProgram) {
			CPACellProgram cmerge = (CPACellProgram) cpa;
			Hashtable cstates = cmerge.stateTable;

			for (Enumeration e = cstates.elements(); e.hasMoreElements();) {
				CellState cs = (CellState) e.nextElement();
				addState(cs);
			}
			firstStates.addAll(cpa.getFirstStates());
		} else {
			Assert.badBranch("Cannot extend: " + cpa.getClass());
		}
	}

	public void addState(CellState p_state) {
		if (firstStates.size() == 0) {
			firstStates.add(p_state.getState());
		}

		if (stateTable.contains(p_state.getState())) {
			Assert.badBranch("Duplicate state: " + p_state);
		} else {
			stateTable.put(p_state.getState(), p_state);
		}
	}

	public void addChemical(ChemicalDescription p_chem) {
		Report.message(0, "adding chemical: " + p_chem);
		if (getChemical(p_chem.getName()) != null) {
			Assert.badBranch("Duplicate chemical: " + p_chem);
		} else {
			chemicalTable.put(p_chem.getName(), p_chem);
			Assert.check(getChemical(p_chem.getName()) == p_chem);
		}
	}

	void resolveStates() {
	}

	public void runCellProgramForOneStep(Cell cell) {
		CellProgram prog = cell.getCellProgram();

		//get the randomly selected state
		CellState cs = getState(cell.getCurrentProgramState());

		if (cs != null) {
			CellTransitionList ctl = cs.getCellTransitionList();

			if (ctl != null) {
				Enumeration enum = ctl.getList().elements();

				while (enum.hasMoreElements()) {
					CellTransition ct = (CellTransition) enum.nextElement();
					ChemicalConditionList ccl = ct.getChemicalConditionList();
					Bindings bindings = ccl.getVariables();
					// Report.message ("Bindings: " + bindings);
					Assert.check(ccl != null);
					if (cell.isConditionSatisfied(ccl, bindings)) {
						Report.message(
							"Condition is satisfied: "
								+ ccl
								+ " / "
								+ bindings);
						// bindings - resolve sets into actual values
						bindings.selectDirections();
						performTransition(cell, ct, bindings);
						break; // only perform first transition
					}
				}
			}
		}
	}

	void performTransition(Cell cell, CellTransition ct, Bindings bindings) {
		Vector argVector = new Vector();
		Report("Performing transition: " + cell + " ==> " + ct);

		if (ct.isDeath()) {
			cell.cellDie(
				bindings,
				ct.getCellEmissionList(),
				ct.getCellDiffusionList());
		} else if (ct.isDivision()) {
			CellDivision cd = ct.getCellDivision();
			Assert.check(cd != null);
			DirectionNode dnode = cd.getDirection();
			if (dnode == null) {
				Report.fatalError(
					"Cell division must have direction: " + cell + ": " + ct);
			}
			Assert.check(dnode != null);
			Direction direction;
			try {
				direction = dnode.eval(bindings);
			} catch (InvalidDirectionException e) {
				Assert.badBranch();
				direction = null;
			}

			Assert.check(direction != null);
			cell.performCellDivision(
				ct.getCellDivision().getParameters1(),
				ct.getCellDivision().getParameters2(),
				direction,
				bindings,
				ct.getCellEmissionList(),
				ct.getCellDiffusionList());
		} else {
			Assert.check(ct.isStateChange());

			cell.changeState(
				ct.getCellStateChange().getParameters(),
				bindings,
				ct.getCellEmissionList(),
				ct.getCellDiffusionList());

		}
	}

	public FColor getColor(Cell c) {
		CellState cs = getState(c.getCurrentProgramState());
		CPAColor color = cs.getColor();

		if (color == null) {
			return new FColor(0, 0, 0);
		}

		return color.eval(c, null);
	}

	public FColor getColor(Cell c, int stateDim) {
		CellState cs = getState(c.getStates()[stateDim]);
		CPAColor color = cs.getColor();

		if (color == null) {
			return new FColor(0, 0, 0);
		}

		return color.eval(c, null);
	}

	public Vector getEmissions(Cell c) {
		CellState cs = (CellState) stateTable.get(c.getCurrentProgramState());
		return cs.getCellEmissionList(c, this);
	}

	public Vector getForwards(Cell c) {
		CellState cs = (CellState) stateTable.get(c.getCurrentProgramState());
		return cs.getCellForwardsList(c);
	}

	public Vector getDiffusions(Cell c) {
		CellState cs = (CellState) stateTable.get(c.getCurrentProgramState());
		return cs.getCellDiffusionList(c, this);
	}

	public Vector provideService(Vector args) {
		Assert.badBranch("Unprovided service: " + args.elementAt(0));
		return null;
	}

	public void init(Vector args) {
	}

	public String toString() {
		StringBuffer retVal = new StringBuffer();
		Enumeration enum = stateTable.elements();

		while (enum.hasMoreElements()) {
			retVal.append(((CellState) enum.nextElement()).toString());
		}
		return retVal.toString();
	}

	public String unparse() {
		StringBuffer retVal = new StringBuffer();
		Enumeration enum = stateTable.elements();

		while (enum.hasMoreElements()) {
			retVal.append(((CellState) enum.nextElement()).unparse());
		}
		return retVal.toString();
	}

	public Vector getFirstStates() {
		if (firstStates == null || firstStates.size() == 0) {
			Assert.badBranch("No states defined!");
		}

		return firstStates;
	}
}
