package parsers.CPA;

import java.util.Enumeration;

import simulator.Bindings;
import simulator.Cell;
import util.*;

public class ChemicalCondition {
	private Term lhs, rhs;
	private Token operator;

	public ChemicalCondition(ChemicalCondition cc) {
		lhs = cc.lhs;
		rhs = cc.rhs;
		operator = cc.operator;
	}

	public ChemicalCondition(Term p_chem1, Token p_operator, Term p_chem2) {
		lhs = p_chem1;
		operator = p_operator;
		rhs = p_chem2;
	}

	public boolean eval(Cell c, Bindings bindings) throws InvalidDirectionException {
		// For each variable, we need to try all possible values, and keep only the true results.
		// Report.message(0, "Term: " + lhs);
		return evalAll(c, bindings, 0);
	}

	private boolean evalAll(Cell c, Bindings bindings, int bindingno)
		{
		boolean res;
		Report.message("evalAll: " + c + " / " + bindings + " / " + bindingno);

		if (bindings == null || bindingno >= bindings.size()) {
			Report.message("Eval Once: " + bindings);
			try {
				res = evalOnce(c, bindings);
			} catch (InvalidDirectionException e) {
				res = false;
			}
		} else {
			res = false;
			// Try all possible settings for the first variable:
			String first = bindings.getKey(bindingno);
			Object value = bindings.lookup(first);
			DirectionSet dirs;
			DirectionSet satisfies = new DirectionSet();

			if (value instanceof Direction) {

				Direction dir = (Direction) value;
				// Report.message(0, "Binding direction: " + dir);
				if (dir.isAny()) {
					// Report.message (0, "Num dimensions: " + c.numDimensions ());
					dirs = Direction.allPossibleDirections(c.numDimensions());
				} else {
					Assert.check(!dir.isSpecialDirection());
					dirs = new DirectionSet(dir);
				}
			} else {
				Assert.check(value instanceof DirectionSet);
				dirs = (DirectionSet) value;
			}

			// Report.message(0, "Binding: " + first + " => " + dirs);
			if (lhs instanceof GoalTerm) {
				Direction bestdir = Direction.UNDEFINED;
				GoalTerm gterm = (GoalTerm) lhs;
				double bestval = 0.0;
				for (Enumeration e = dirs.elements(); e.hasMoreElements();) {
					try {
						Direction dir = (Direction) e.nextElement();
						// Report.message (0, "Trying direction: " + dir);
						bindings.updateBinding(first, dir);
						double lval = gterm.eval(c, bindings);

						if (bestdir.isUndefined()
							|| (gterm.isMinimize() && lval < bestval)
							|| (gterm.isMaximize() && lval > bestval)) {
							bestdir = dir;
							bestval = lval;
						}
					} catch (InvalidDirectionException ide) {
						// Report.message (0, "Invalid!");
						// Okay, try the next direction (attempted direction was not in the direction set for some binding
					}
				}

				bindings.updateBinding(first, bestdir);

				if (evalAll(c, bindings, bindingno + 1)) {
					satisfies.add(bestdir);
					Report.message("Satisfied for dir: " + bestdir);
					res = true;
				} else {
					Report.message("Not satisfied for dir: " + bestdir);
				}
			} else {
				for (Enumeration e = dirs.elements(); e.hasMoreElements();) {
					Direction dir = (Direction) e.nextElement();
					bindings.updateBinding(first, dir);
					if (evalAll(c, bindings, bindingno + 1)) {
						satisfies.add(dir);
						// Report.message ("Satisfied for dir: " + dir);
						res = true;
					} else {
						//	Report.message ("Not satisfied for dir: " + dir);				
					}
				}
			}

			if (res == true) {
				Assert.check(!satisfies.isEmpty());
			} else {
				Assert.check(satisfies.isEmpty());
			}

			// Replace the binding with only those values that satisfy the condition
			bindings.updateBinding(first, satisfies);
		}

		// Report.message (0, "Returning " + res + " / " + bindings);
		return res;
	}

	private boolean evalOnce(Cell c, Bindings bindings)
		throws InvalidDirectionException {
		double lval = lhs.eval(c, bindings);
		// Report.message("evaluating lhs: " + lhs + " -> " + lval);
		double rval = rhs.eval(c, bindings);
		// Report.message("evaluating rhs: " + rhs + " -> " + rval);
		boolean res;

		switch (operator.kind) {
			case CPAParserConstants.EQUALS :
				res = (lval == rval);
				break;
			case CPAParserConstants.LT :
				res = (lval < rval);
				break;
			case CPAParserConstants.GT :
				res = (lval > rval);
				break;
			default :
				Assert.badBranch("Bad operator: " + operator.image);
				res = false;
		}

		return res;
	}

	public Bindings getVariables() {
		Bindings lbindings = lhs.getVariables();
		Bindings rbindings = rhs.getVariables();

		if (lbindings != null) {
			lbindings.addAll(rbindings);
			return lbindings;
		} else {
			return rbindings;
		}
	}

	public String toString() {
		return "("
			+ lhs.toString()
			+ " "
			+ operator.image
			+ " "
			+ rhs.toString()
			+ ")";
	}
}
