package simulator;

import java.util.*;

import util.*;
import parsers.CPA.*;
import parsers.CPA.CPACellProgram;

public class Cell {
	private boolean isNewCell;
	private boolean hadSerious;
	private int id;
	private SystemModel sys;

	//multiple program code
	private CPACellProgram cp;
	private String[] states;
	private int currentProgram;

	// Note that the sensor bins are stored according to the 
	// absolute direction!
	private Vector sensedChemicalsDir[];

	private Vector sensedChemicals;
	private Vector chemicalsToEmit;
	private Vector chemicalsToStore;
	private Vector chemicalsToDiffuse;
	protected Position position;

	private Orientation orientation;
	private Bindings variables;

	protected void Report(String s) {
		Report.message(
			"[Cell "
				+ getID()
				+ "/"
				+ currentProgram
				+ "/"
				+ states[currentProgram]
				+ "/"
				+ position
				+ "] "
				+ s);
	}

	private void Report(int i, String s) {
		Report.message(
			i,
			"[Cell "
				+ getID()
				+ "/"
				+ currentProgram
				+ "/"
				+ states[currentProgram]
				+ "/"
				+ position
				+ "] "
				+ s);
	}

	protected void ReportSerious(String s) {
		Report.message(
			0,
			"[Cell "
				+ getID()
				+ "/"
				+ currentProgram
				+ "/"
				+ states[currentProgram]
				+ "/"
				+ position
				+ "] "
				+ s);
		hadSerious = true;
	}

	public boolean hadSerious() {
		return hadSerious;
	}

	public Cell(
		int p_id,
		SystemModel p_sys,
		CellProgram p_cp,
		Vector p_states,
		int programNum) {

		id = p_id;
		sys = p_sys;
		currentProgram = programNum;

		//save memory by all cells pointing to only 1 cell program
		cp = (CPACellProgram) p_cp;

		//copy the current states from the parent
		states = new String[p_states.size()];
		//System.arraycopy(p_states, 0, states, 0, p_states.length);
		p_states.toArray(states);

		sensedChemicals = new Vector();
		sensedChemicalsDir =
			new Vector[Direction.numDirections(numDimensions())];
		for (int i = 0; i < Direction.numDirections(numDimensions()); i++) {
			sensedChemicalsDir[i] = new Vector();
		}
		chemicalsToEmit = new Vector();
		chemicalsToStore = null; // not implemented!
		chemicalsToDiffuse = new Vector();
		variables = new Bindings();
		orientation =
			new Orientation(
				Math.toRadians(0.0),
				Math.toRadians(0.0),
				Math.toRadians(0.0));
	}

	public int getID() {
		return id;
	}

	public int getCurrentProgramNum() {
		return currentProgram;
	}

	public String getCurrentProgramState() {
		return states[currentProgram];
	}

	public void setCurrentProgramState(String p_state) {
		states[currentProgram] = p_state;
	}
	
	public String[] getStates() {
		return states;
	}

	public void setStates(String[] p_states) {
		if (p_states.length > states.length) {
			System.err.print("p_states is longer than states");
			return;
		}

		System.arraycopy(p_states, 0, states, 0, p_states.length);
	}

	public Vector getSensor(Direction dir) {
		return sensedChemicalsDir[dir.directionIndex()];
	}

	public SystemModel getSystem() {
		return sys;
	}
	
	//name isn't very good since there is only one "program" per cell now
	public int getNumPrograms() {
		return states.length;
	}

	public CellProgram getCellProgram() {
		return cp;
	}

	public void runCellProgramForOneStep() {
		if (states.length > 1) {
			currentProgram = SimRandom.nextInt(states.length);
		}

		cp.runCellProgramForOneStep(this);
	}

	public void disposeSensedChemicals() {
		sensedChemicals.clear();
		for (int i = 0; i < Direction.numDirections(numDimensions()); i++) {
			sensedChemicalsDir[i].clear();
		}
	}

	public FColor getColor() {
		return new FColor(1.0, 1.0, 1.0);
	}

	public boolean isNewCell() {
		return isNewCell;
	}

	/**
	 * @param b
	 */
	public void setNewCell(boolean b) {
		isNewCell = b;
	}

	public Position getPosition() {
		return position;
	}

	/**
	 * @return
	 */
	public double getX() {
		return position.getX();
	}

	/**
	 * @return
	 */
	public double getY() {
		return position.getY();
	}

	/**
	 * @return
	 */
	public double getZ() {
		return position.getZ();
	}

	public void setPosition(double p_x, double p_y, double p_z) {
		position = new Position(p_x, p_y, p_z);
	}

	public void setPosition(int p_x, int p_y, int p_z) {
		position = new Position(p_x, p_y, p_z);
	}

	public void setPosition(Position p_pos) {
		position = p_pos;
	}

	// This is dangerous, but the sysmodel updates chemical concentrations this way
	/*@exposed@*/
	public Vector getSensedChemicals() {
		return sensedChemicals;
	}

	protected /*@exposed@*/
	Vector getSensedChemicalsFromDirection(Direction dir) {
		// The direction is an absolute direction.  So,
		// we need to orient it.

		if (dir == null) {
			Assert.badBranch();
			return sensedChemicals;
		}

		if (dir.isAny() || dir.isAll() || dir.isUndefined()) {
			return sensedChemicals;
		}

		Direction actualdir = dir.orient(orientation);
		return sensedChemicalsDir[actualdir.directionIndex()];
	}

	private ChemicalConcentration getChemicalBin(String chem) {
		for (Enumeration enum = sensedChemicals.elements();
			enum.hasMoreElements();
			) {
			ChemicalConcentration cc =
				(ChemicalConcentration) enum.nextElement();
			if (cc.getChemical().equals(chem)) {
				return cc;
			}
		}

		return null;
	}

	public void executeStateSecretions() {
		/*
		 * 	Add the state secretions from the cell program.  The chemicalsToEmit
		 * and chemicalsToDiffuse variables contain the transition chemicals already.
		 */

		// Report ("< executing state secretions >");
		Vector retVec = getCellProgram().getEmissions(this);
		if (retVec != null) {
			Report("Adding emissions: " + retVec);
			chemicalsToEmit.addAll(retVec);
		}

		retVec = getCellProgram().getDiffusions(this);

		if (retVec != null) {
			chemicalsToDiffuse.addAll(retVec);
		}

		if (chemicalsToEmit != null && chemicalsToEmit.size() > 0) {
			getSystem().emitChemicals(this, chemicalsToEmit);
			chemicalsToEmit.clear();
		}

		if (chemicalsToDiffuse != null && chemicalsToDiffuse.size() > 0) {
			getSystem().diffuseChemicals(this, chemicalsToDiffuse);
			chemicalsToDiffuse.clear();
		}
	}

	// If the condition is satisfied, returns a Bindings that satisfies it.
	// Otherwise, returns null.
	//
	// For now, we evaluate each condition (left to right), and fix the bindings.
	// Hence, pattern matching between conditions will not explore all possible bindings.

	public boolean isConditionSatisfied(
		ChemicalConditionList condition,
		Bindings bindings) {
		Vector chemConditionsVector = condition.getList();
		boolean conditionSatisfied = true;

		Report("isConditionSatisfied: " + condition + " bindings: " + bindings);
		Enumeration enum = chemConditionsVector.elements();

		while (conditionSatisfied && enum.hasMoreElements()) {
			ChemicalCondition chemCondition =
				(ChemicalCondition) enum.nextElement();
			Report("Condition: " + chemCondition + " bindings: " + bindings);
			try {
				if (!chemCondition.eval(this, bindings)) {
					conditionSatisfied = false;
					Report("Not satisfied: " + chemCondition);
				}
			} catch (InvalidDirectionException e) {
				// One of the bindings directions is invalid
				conditionSatisfied = false;
				Report(0, "Invalid direction: " + e);
			}
		}

		Report(
			"Condition "
				+ condition
				+ " satisfied? "
				+ conditionSatisfied
				+ " bindings: "
				+ bindings);

		return conditionSatisfied;
	}

	/*
	 * Cell devides into 2 cells in 2 new state.
	 * The emissions and diffusions are secreted by the original 
	 * cell (which becomes the first of the two daughter cells)
	 * at the *next* step.  So, we need to store the chemicals to 
	 * secrete until the next time executeStateSecretions is called.
	 */
	public void performCellDivision(
		CellParameters params1,
		CellParameters params2,
		Direction divisionDirection,
		Bindings bindings,
		CellSecretionNodeList emissions,
		CellSecretionNodeList diffusions) {
		recordSecretions(emissions, diffusions, bindings);
		Report(
			3,
			"Direction before orientation ["
				+ orientation
				+ "]: "
				+ divisionDirection);
		Direction absoluteDirection = divisionDirection.orient(orientation);
		Report(3, "Direction after orientation: " + absoluteDirection);

		getSystem().performCellDivision(
			this,
			bindings,
			params1,
			params2,
			absoluteDirection);
		Report("*** After Division ***");
	}

	/**
	 * @param emissions
	 * @param diffusions
	 * @param sensedDirection
	 */
	private void recordSecretions(
		CellSecretionNodeList emissions,
		CellSecretionNodeList diffusions,
		Bindings bindings) {
		if (emissions != null) {
			recordEmissions(emissions.getList(), bindings);
		}

		if (diffusions != null) {
			recordDiffusions(diffusions.getList(), bindings);
		}
	}

	/**
	 * @param vector
	 * @param divisionDirection
	 * @param p_givenDirection
	 */
	private void recordDiffusions(Vector vector, Bindings sensedDirection) {
		Assert.badBranch("Cannot do diffusions!");
	}

	/**
	 * @param vector
	 * @param divisionDirection
	 * @param p_givenDirection
	 */
	private void recordEmissions(Vector emits, Bindings bindings) {
		Report("Recording emissions: " + emits);
		for (Enumeration enum = emits.elements(); enum.hasMoreElements();) {
			CellEmissionNode emit = (CellEmissionNode) enum.nextElement();
			DirectionNode dirnode = emit.getDirection();
			Direction dir;

			if (dirnode == null) {
				dir = Direction.ALL;
			} else {
				try {
					dir = dirnode.eval(bindings);
				} catch (InvalidDirectionException e) {
					Assert.badBranch();
					dir = null;
				}

				if (!dir.isSpecialDirection()) {
					Direction actualdir = dir.orient(orientation);
					// System.err.println("Emit: " + dir + " ==> " + actualdir);
					dir = actualdir;
				}
			}

			try {
				recordEmission(
					new CellEmission(
						emit.getChemical(),
						emit.getConcentration().eval(this, bindings),
						dir));
			} catch (InvalidDirectionException e) {
				Assert.badBranch();
			}
		}
	}

	/**
	 * @param string
	 * @param sensedDirection
	 * @return
	 */

	public double lookupSensedConcentration(String chem) {
		return lookupSensedConcentration(chem, Direction.UNDEFINED);
	}

	public double lookupSensedConcentration(
		String chem,
		Direction sensedDirection) {
		double envconc = 0.0;

		Report("Looking up: " + chem + " from " + sensedDirection);
		if (sensedDirection.isUndefined()
			|| sensedDirection.isAny()
			|| sensedDirection.isAll()) {
			envconc =
				sys.getConcentration(
					getPosition(),
					chem,
					sys.getSimulation().getStep());
		} else {
			Position pos =
				getPosition().add(sensedDirection.canonicalize().scale(0.5));
			envconc =
				sys.getConcentration(pos, chem, sys.getSimulation().getStep());
		}

		Vector chems = getSensedChemicalsFromDirection(sensedDirection);

		for (Enumeration cenum = chems.elements(); cenum.hasMoreElements();) {
			ChemicalConcentration cemit =
				(ChemicalConcentration) cenum.nextElement();

			if (cemit.getChemical().equals(chem)) {
				return cemit.getConcentration() + envconc;
			}
		}

		return envconc;
	}

	/**
	 * @param ce
	 * @param p_givenDirection
	 */
	private void recordEmission(CellSecretion secretion) {
		recordSecretion(chemicalsToEmit, secretion);
	}

	/**
	 * @param chemicalsToEmit
	 * @param ce
	 * @param p_givenDirection
	 */
	private void recordSecretion(Vector chemicalList, CellSecretion ce) {
		Assert.check(
			chemicalList == chemicalsToEmit
				|| chemicalList == chemicalsToDiffuse);
		if (ce.getConcentration() > 0.0) {
			chemicalList.add(ce);
		}
	}

	/**
	 * @param string
	 * @param sensedDirection
	 * @param list
	 * @param list2
	 */
	public void changeState(
		CellParameters params,
		Bindings bindings,
		CellSecretionNodeList emissions,
		CellSecretionNodeList diffusions) {
		setCurrentProgramState(params.getState());
		reflectParameters(this, bindings, params);
		recordSecretions(emissions, diffusions, bindings);
	}

	public void cellDie(
		Bindings bindings,
		CellSecretionNodeList emissions,
		CellSecretionNodeList diffusions) {
		showChemicals();
		Report("Cell Die");
		recordSecretions(emissions, diffusions, bindings);
		// will this even happen?!
		getSystem().cellSuicide(this);
	}

	private void addSensedChemicals(Vector sensor, Vector emissions) {
		for (Enumeration enum = emissions.elements();
			enum.hasMoreElements();
			) {
			CellEmission emission = (CellEmission) enum.nextElement();
			addSensedChemical(sensor, emission);
		}
	}

	private void addSensedChemical(Vector sensor, CellEmission emission) {
		if (sensor == null)
			return;
		// Report("adding sensed chemical: " + emission);
		Enumeration enum = sensor.elements();
		while (enum.hasMoreElements()) {
			ChemicalConcentration cc =
				(ChemicalConcentration) enum.nextElement();
			if (emission.getChemical().equals(cc.getChemical())) {
				cc.setConcentration(
					cc.getConcentration() + emission.getConcentration());
				return;
			}
		}

		// Not in sensor vector, add a new entry
		sensor.add(
			new ChemicalConcentration(
				emission.getChemical(),
				emission.getConcentration()));
	}

	/*
	 * addSensedChemical
	 * 
	 * This is called when another cell executes a diffuse.  Once the environment
	 * model is ready, it will be removed. 
	 */

	public void addSensedChemical(CellEmission emission) {
		// The concentration is the max of all emissions of a chemical
		// (why isn't it the sum?)
		addSensedChemical(sensedChemicals, emission);
	}

	public void receiveChemical(Cell cd, CellEmission emission) {
		// The bin that senses the emission depends on the direction of
		// the sending cell.
		Report("Received chemical: " + emission);

		if (getX() > cd.getX())
			addSensedChemical(getSensor(Direction.XM), emission);
		if (getX() < cd.getX())
			addSensedChemical(getSensor(Direction.XP), emission);

		if (numDimensions() > 1) {
			if (getY() > cd.getY())
				addSensedChemical(getSensor(Direction.YM), emission);
			if (getY() < cd.getY())
				addSensedChemical(getSensor(Direction.YP), emission);
			if (numDimensions() > 2) {
				if (getZ() > cd.getZ())
					addSensedChemical(getSensor(Direction.ZM), emission);
				if (getZ() < cd.getZ())
					addSensedChemical(getSensor(Direction.ZP), emission);
			}
		}

		addSensedChemical(emission);
		showChemicals();
	}

	public String toString() {
		return "[Cell "
			+ getID()
			+ "/"
			+ currentProgram
			+ "/"
			+ states[currentProgram]
			+ "]";
	}

	private void showChemicals() {
		Report("Chemicals: " + sensedChemicals);
		/* can't show all bins, depends on dimensions!				
					+ " / X-: "
					+ getSensor(Direction.XM)
					+ " X+: "
					+ getSensor(Direction.XP)
					+ "Y-: "
					+ getSensor(Direction.YM)
					+ "Y+: "
					+ getSensor(Direction.YP)
					+ "Z-: "
					+ getSensor(Direction.ZM)
					+ "Z+: "
					+ getSensor(Direction.ZP));
					*/
	}

	public String display() {
		StringBuffer sb = new StringBuffer();

		for (int i = 0; i < Direction.numDirections(numDimensions()); i++) {
			Enumeration enum = sensedChemicalsDir[i].elements();
			sb.append(" | " + Direction.indexDirection(i).toString() + " ");

			while (enum.hasMoreElements()) {
				ChemicalConcentration cc =
					(ChemicalConcentration) enum.nextElement();
				sb.append(cc.toString() + ",");
			}
		}

		sb.append(" | " + Direction.ALL.toString() + " ");

		Enumeration enum = sensedChemicals.elements();
		while (enum.hasMoreElements()) {
			ChemicalConcentration cc =
				(ChemicalConcentration) enum.nextElement();
			sb.append(cc.toString() + ",");
		}

		sb.append("\n");
		return sb.toString();
	}

	public int numDimensions() {
		return 3;
	}

	public void reflectParameters(
		Cell cell,
		Bindings bindings,
		CellParameters params) {
		// Change the cell state according to the parameters
		// cell is the source cell (for evaluating bindings)
		Bindings vars = params.getParameters();

		// The initial orientation is the same as the old cell
		if (this != cell) {
			orientation = new Orientation(cell.orientation);
		}

		if (vars != null) {
			for (Enumeration enum = vars.getKeys(); enum.hasMoreElements();) {
				String key = (String) enum.nextElement();

				if (key.equals("orientation")) {
					Term term = (Term) vars.lookup(key);
					Report(
						3,
						"Setting orientation: "
							+ orientation
							+ " rotate "
							+ term);
					Direction dir;

					try {
						dir = term.evalDir(cell, bindings);
					} catch (InvalidDirectionException e) {
						Assert.badBranch();
						dir = null;
					}

					if (dir == null) {
						Report(
							0,
							"No direction: " + term + " with " + bindings);
					} else {
						dir = dir.orient(cell.orientation);
						// The direction should be relative to the cell's orientation
						orientation.add(dir.canonicalize());
						Report(3, "New orientation: " + orientation);
					}
				} else {
					Report.message(0, "Set binding for: " + key);
					Term term = (Term) vars.lookup(key);

					double val;
					try {
						val = term.eval(cell, bindings);
					} catch (InvalidDirectionException e) {
						Assert.badBranch();
						val = 0.0;
					}

					Report(
						0,
						"Setting variable: "
							+ key
							+ " ==> "
							+ term
							+ " = "
							+ val);

					variables.addBinding(key, new Double(val));

				}
			}
		}
	}

	/**
	 * @param chemical
	 * @return
	 */
	public double getVariableValue(String chemical) {
		double tmp = ((Double) variables.lookup(chemical)).doubleValue();
		return tmp;
	}

	/**
	 * @param chemical
	 * @return
	 */
	public boolean hasVariable(String chemical) {
		return (variables.lookup(chemical) != null);
	}

	public boolean matchesState(String state) {
		for (int i = 0; i < getNumPrograms(); i++) {
			if (states[i].equals(state)) {
				return true;
			}
		}
		return false;
	}
}