package simulator;

import java.util.*;
import java.io.*;
import javax.swing.JOptionPane;

import util.*;

public class Simulation {
	public static Hashtable sysModels = new Hashtable();
	public static Hashtable cellModels = new Hashtable();
	public static Hashtable parsers = new Hashtable();

	public static int dnaDefault = 3;
	private static final String DEFAULT_CELLMODEL =
		"cellmodels.simplecell.SimpleCell";
	private static final String DEFAULT_PARSER = "parsers.CPA.CPAParser";

	private Properties configuration = new Properties();
	private SimFileTokenizer simFileTokenizer;
	private BufferedWriter outputWriter;
	private BufferedWriter statsWriter;
	private BufferedWriter errorWriter;

	private String systemModelName;
	private String systemModelOptions;
	private String cellModelName;
	private String cellModelOptions;
	private String parserModelName;
	private String parserModelOptions;
	private String title;
	private SimToken nextSimToken;

	private SystemModel sys;
	//private CellProgram cellProgram;
	private CellProgram cellProgram = null;
	private SimulationParameters params;
	private int step = 0;

	public Simulation(SimulationParameters p_params) {
		params = p_params;

		String errFileName = params.getErrFile();

		if (errFileName != null) {
			try {
				errorWriter = new BufferedWriter(new FileWriter(errFileName));
			} catch (FileNotFoundException e) {
				fatalError(
					"Unable to find error output file "
						+ errFileName
						+ ": "
						+ e.toString());
				return;
			} catch (IOException e) {
				fatalError(
					"Unable to open error output file "
						+ errFileName
						+ ": "
						+ e.toString());
				return;
			}
		} else {
			errorWriter = new BufferedWriter(new PrintWriter(System.out));
		}

		if (params.hasRandomSeed()) {
			SimRandom.init(params.getRandomSeed());
		} else {
			SimRandom.init();
		}

		/*
		** Set DNA Default Value
		*/
		if (!(params.getDNAinit() == 0)) {
			dnaDefault = params.getDNAinit();
		}

		boolean nosimfile = false;

		/*
		** Read sim file
		*/
		try {
			String simFileName = params.getSimFile();

			if (simFileName != null) {
				title = simFileName + " " + params.getProgramFile();
				BufferedReader simFileStream =
					new BufferedReader(new FileReader(simFileName));
				simFileTokenizer = new SimFileTokenizer(simFileStream);
			} else {
				title = params.getProgramFile();
				nosimfile = true;
			}
		} catch (FileNotFoundException e) {
			fatalError(
				"Unable to find sim file "
					+ params.getSimFile()
					+ ": "
					+ e.toString());
			return;
		}

		try {
			boolean sysModelRead = false,
				cellModelRead = false,
				parserModelRead = false;

			if (simFileTokenizer != null) {
				nextSimToken = simFileTokenizer.nextSimToken();

				while ((nextSimToken.getType() == SimToken.SIM_CONFIG)
					&& (!sysModelRead || !cellModelRead || !parserModelRead)) {
					if (nextSimToken.getType() == SimToken.SIM_CONFIG) {
						if (nextSimToken.getConfigParam().equals("system")) {
							params.setSystemModel(
								nextSimToken.getConfigValue());
							sysModelRead = true;
						} else if (
							nextSimToken.getConfigParam().equals("cell")) {
							configuration.setProperty(
								"cell",
								nextSimToken.getConfigValue());
							cellModelRead = true;
						} else if (
							nextSimToken.getConfigParam().equals("parser")) {
							Assert.badBranch("Don't support parser models!");
							/* configuration.setProperty(
								"parser",
								nextSimToken.getConfigValue());
							parserModelRead = true;						
							*/
						}
					}
					nextSimToken = simFileTokenizer.nextSimToken();
				}
			}

			configuration.setProperty("system", params.getSystemModel());

			if (!cellModelRead) {
				configuration.setProperty("cell", DEFAULT_CELLMODEL);
			}

			if (!parserModelRead) {
				configuration.setProperty("parser", DEFAULT_PARSER);
			}
		} catch (IOException e) {
			fatalError(
				"Io Exception ocurred reading sim file: " + e.toString());
			System.exit(0);
		}

		systemModelName = new String(configuration.getProperty("system"));
		systemModelOptions = new String("");
		cellModelName = new String(configuration.getProperty("cell"));
		cellModelOptions = new String("");
		parserModelName = new String(configuration.getProperty("parser"));
		parserModelOptions = new String("");

		/*
		** Create output file
		*/

		String outFileName = params.getOutFile();

		if (outFileName != null) {
			try {
				outputWriter = new BufferedWriter(new FileWriter(outFileName));
			} catch (FileNotFoundException e) {
				fatalError(
					"Unable to find output file "
						+ outFileName
						+ ": "
						+ e.toString());
				return;
			} catch (IOException e) {
				fatalError(
					"Unable to open output file "
						+ outFileName
						+ ": "
						+ e.toString());
				return;
			}
		} else {
			outputWriter = new BufferedWriter(new PrintWriter(System.out));
		}

		String statsFileName = params.getStatsFile();

		if (statsFileName != null) {
			try {
				statsWriter = new BufferedWriter(new FileWriter(statsFileName));
			} catch (FileNotFoundException e) {
				fatalError(
					"Unable to find statsput file "
						+ statsFileName
						+ ": "
						+ e.toString());
				return;
			} catch (IOException e) {
				fatalError(
					"Unable to open statsput file "
						+ statsFileName
						+ ": "
						+ e.toString());
				return;
			}
		} else {
			statsWriter = new BufferedWriter(new PrintWriter(System.out));
		}

		Vector cellPrograms_ = params.getCellPrograms();
		for (Enumeration e = cellPrograms_.elements(); e.hasMoreElements();) {
			String cellProgramName = (String) e.nextElement();
			BufferedReader cellProgramReader;

			try {
				cellProgramReader =
					new BufferedReader(new FileReader(cellProgramName));

				////////////////////////////////////////////////
				////////////////////////////////////////////////
				//create cell programs here
				////////////////////////////////////////////////
				////////////////////////////////////////////////

				try {
					if (cellProgram == null)
					{
					
						cellProgram =
							createCellProgram(
								parserModelName,
								parserModelOptions,
								cellProgramReader);
					}
					else
					{
						cellProgram.extend(createCellProgram(parserModelName,
															parserModelOptions,
															cellProgramReader));
					}
				} catch (GenericParseException gpe) {
					Report.fatalError(
						"Parse error in "
							+ cellProgramName
							+ ": "
							+ gpe.toString());
				}

			} catch (FileNotFoundException fnfe) {
				System.err.println(
					"Unable to find cell program file "
						+ cellProgramName
						+ ": "
						+ fnfe.toString());
			}
		}

		try {
			sys = createSystem(systemModelName, systemModelOptions);
		} catch (Exception e) {
			System.err.println("Cannot create system: " + e.toString());
		}

		if (nosimfile) {			
			//always start in first program?
			int firstProgram = 0;

			sys.installNewCell(
				getCellProgram().getFirstStates(),
				new Position(0.0, 0.0, 0.0),
				firstProgram);
		}

		runInstructions();
		sys.updateStatistics();
	}

	public StatsTable getStatistics() {
		return sys.getStatistics();
	}

	public CellProgram getCellProgram() {
		return cellProgram;
	}

	private void runInstructions() {
		if (simFileTokenizer != null) {
			while (nextSimToken.getType() == SimToken.SIM_INSTRUCTION
				&& nextSimToken.getInstructionNumber() <= getStep()) {
				sys.runInstruction(nextSimToken.getInstruction());
				try {
					nextSimToken = simFileTokenizer.nextSimToken();
				} catch (IOException e) {
					System.err.println(e.toString());
				}
			}
		}
	}

	public void runToCompletion() throws IOException {
		int maxsteps = params.getMaxSteps();

		// writeOutputText ("Simulating " + maxsteps + " steps: ");
		while (step < maxsteps) {
			// writeOutputText (".");
			nextStep();
		}
	}

	private void logState() {
		if (params.logCounts()) {
			try {
				writeOutputLine(step + ": " + sys.numberOfCells());
			} catch (IOException ioe) {
				System.err.println("IO Exception: " + ioe);
			}
		}
		if (params.logAll()) {
			try {
				writeOutputLine("=== " + step + " ===");
				sys.generateOutput(outputWriter);
				outputWriter.flush();
			} catch (IOException ioe) {
				System.err.println("IO Exception: " + ioe);
			}
		}
	}

	public void nextStep() {
		if (step == 0) {
			logState();
		}

		step++;
		runInstructions();
		sys.runCellProgramForOneStep();
		logState();
	}

	public int getStep() {
		return step;
	}

	public int getDNADefault() {
		return dnaDefault;
	}

	public String getTitle() {
		return title + ": " + step;
	}

	public void writeOutputLine(String s) throws IOException {
		outputWriter.write(s, 0, s.length());
		outputWriter.newLine();
		outputWriter.flush();
	}

	public void writeOutputText(String s) throws IOException {
		outputWriter.write(s, 0, s.length());
	}

	public void writeStatsLine(String s) throws IOException {
		statsWriter.write(s, 0, s.length());
		statsWriter.newLine();
		statsWriter.flush();
	}

	public void writeErrorLine(String s) {
		try {
			errorWriter.write(s, 0, s.length());
			errorWriter.newLine();
			errorWriter.flush();
		} catch (IOException ioe) {
			System.err.println("Error writing error line: " + s + " / " + ioe);
		}
	}

	public void writeErrorText(String s) {
		try {
			errorWriter.write(s, 0, s.length());
		} catch (IOException ioe) {
			System.err.println("Error writing error line: " + s + " / " + ioe);
		}

	}

	public void finish() {
		try {
			// statsWriter.write("Execution finished: " + getStep () + " steps.");
			if (params.getCSVStats()) {
				writeStatsLine(getStatistics().toCSVString());
			} else {
				// writeStatsLine(getStatistics().toString());
			}

			outputWriter.flush();
			outputWriter.close();
			statsWriter.flush();
			statsWriter.close();
			errorWriter.flush();
			errorWriter.close();
		} catch (IOException e) {
			System.err.println("Error closing files: " + e);
		}
	}

	public Cell createCell(
		int p_ID,
		SystemModel p_sys,
		CellProgram p_cp,
		Vector p_states,
		int programNum) {
		CellFactory s = (CellFactory) cellModels.get(cellModelName);

		if (s == null) {
			try {
				Class.forName(cellModelName);
				s = (CellFactory) cellModels.get(cellModelName);

				if (s == null) {
					Assert.badBranch(
						"Could not create cell class: " + cellModelName);
				}
			} catch (ClassNotFoundException e) {
				Assert.badBranch("Could not find cell class: " + cellModelName);
			} catch (NoClassDefFoundError e) {
				Assert.badBranch("Error loading cell class: " + cellModelName);
			}
		}

		return s.create(
			p_ID,
			p_sys,
			p_cp,
			cellModelOptions,
			p_states,
			programNum);
	}

	public CellProgram createCellProgram(
		String modelName,
		String modelOptions,
		BufferedReader inp)
		throws GenericParseException {
		ParserFactory s = (ParserFactory) parsers.get(modelName);

		if (s == null) {
			try {
				Class.forName(modelName);
				s = (ParserFactory) parsers.get(modelName);
				if (s == null) {
					Assert.badBranch("Could not find parser: " + modelName);
				}
			} catch (ClassNotFoundException e) {
				Assert.badBranch("Could not find parser class: " + modelName);
			} catch (NoClassDefFoundError e) {
				Assert.badBranch("Error loading parser class: " + modelName);
			}
		}

		Parser p = s.create(inp, modelOptions);

		try {
			return p.getCellProgram();
		} catch (GenericParseException e) {
			throw new GenericParseException(e.toString());
		}
	}

	public SystemModel createSystem(String modelName, String modelOptions) {
		SystemFactory s = (SystemFactory) sysModels.get(modelName);

		if (s == null) {
			try {
				Class.forName(modelName);
				// By now the static block with no function would
				// have been executed if the System was found.
				// the System is expected to have put its factory
				// in the hashtable.
				s = (SystemFactory) sysModels.get(modelName);

				if (s == null) {
					Assert.badBranch("Could not create system: " + modelName);
				}
			} catch (ClassNotFoundException e) {
				Assert.badBranch("Could not find system class: " + modelName);
			} catch (NoClassDefFoundError e) {
				Assert.badBranch("Error loading system class: " + modelName);
			} catch (Exception e) {
				Assert.badBranch("Unhandled Exception creating system: " + e);
			}
		}

		return s.create(this, modelOptions);
	}

	public void fatalError(String s) {
		if (params.isBatchMode()) {
			System.err.println(s);
			System.exit(1);
		} else {
			JOptionPane.showMessageDialog(
				null,
				s,
				"Fatal error",
				JOptionPane.ERROR_MESSAGE);
			System.exit(1);
		}

	}

	public SystemModel getSystemModel() {
		Assert.check(sys != null);
		return sys;
	}

	public double getDither() {
		return params.getDither();
	}

	public boolean getAutoScale() {
		return params.getAutoScale();
	}
}
