package simulator;

import util.Assert;
import java.util.Vector;

public class SimulationParameters {

	private static final String DEFAULT_SYSTEMMODEL =
		"sysmodels.standard.StandardModel";
	private String systemModel;
	private boolean compactmode;
	private Vector cpaFiles; // vector of String's
	private String simFile;
	private String errFile;
	private String statsFile;
	private String outFile;
	private boolean batchMode = false;
	private int maxSteps;
	private boolean csvStats = false;
	private boolean hasRandomSeed = false;
	private long randomSeed;
	private int dnaValue;
	private double dither = 0.0;
	private boolean autoscale = true;
	private boolean logall = false;
	private boolean logcounts = false;
	private boolean warnUndeclaredChemicals = true;
	private boolean isDisplay3D = true;

	static private SimulationParameters simparams;

	public SimulationParameters() {
		cpaFiles = new Vector();
		simparams = this; // Only one SimulationParameters object
		systemModel = DEFAULT_SYSTEMMODEL;
	}

	static public SimulationParameters getSimulationParameters() {
		return simparams;
	}

	public void setRandomSeed(long val) {
		randomSeed = val;
		hasRandomSeed = true;
	}

	public boolean hasRandomSeed() {
		return hasRandomSeed;
	}

	public long getRandomSeed() {
		Assert.check(hasRandomSeed);
		return randomSeed;
	}

	public void setSimFile(String sim) {
		Assert.check(simFile == null);
		simFile = sim;
	}

	public void setErrFile(String err) {
		Assert.check(errFile == null);
		errFile = err;
	}

	public void setDNAinit(int dnaIni) {
		Assert.check(dnaIni >= 0);
		dnaValue = dnaIni;
	}

	public void setOutFile(String out) {
		Assert.check(outFile == null);
		outFile = out;
	}

	public void setStatsFile(String stats) {
		Assert.check(statsFile == null);
		statsFile = stats;
	}

	public void setBatchMode(boolean mode) {
		batchMode = mode;
	}

	public void setCSVStats(boolean val) {
		csvStats = val;
	}

	public boolean isBatchMode() {
		return batchMode;
	}

	public void setMaxSteps(int steps) {
		maxSteps = steps;
	}

	public void addProgramFile(String cpa) {
		cpaFiles.add(cpa);
	}

	public void setSystemModel(String s) {
		systemModel = s;
	}

	public String getSystemModel() {
		return systemModel;
	}
	public boolean parametersDefined() {
		return cpaFiles.size() > 0;
	}

	public String getSimFile() {
		return simFile;
	}
	public String getOutFile() {
		return outFile;
	}
	public int getDNAinit() {
		return dnaValue;
	}
	public String getErrFile() {
		return errFile;
	}
	public boolean getCSVStats() {
		return csvStats;
	}
	public String getStatsFile() {
		return statsFile;
	}
	public int getMaxSteps() {
		return maxSteps;
	}
	public Vector getCellPrograms() {
		return cpaFiles;
	}
	public String getProgramFile() {
		if (cpaFiles.size() > 0)
			return (String) cpaFiles.elementAt(0);
		else
			return null;
	}

	public void setDither(double d) {
		dither = d;
	}

	public double getDither() {
		return dither;
	}

	public void setNoScale() {
		autoscale = false;
	}

	public boolean getAutoScale() {
		return autoscale;
	}

	public void setLogAll() {
		logall = true;
	}

	public boolean logAll() {
		return logall;
	}

	public void setLogCounts() {
		logcounts = true;
	}

	public boolean logCounts() {
		return logcounts;
	}

	public void setCompactMode(boolean b) {
		compactmode = b;
	}

	public boolean isCompact() {
		return compactmode;
	}

	public void setWarnUndeclaredChemicals() {
		warnUndeclaredChemicals = true;
	}

	public void setNoWarnUndeclaredChemicals() {
		warnUndeclaredChemicals = false;
	}

	public boolean warnUndeclaredChemicals() {
		return warnUndeclaredChemicals;
	}

	private static boolean prefixEquals(String param, String s) {
		if (param.length() >= s.length()) {
			return (param.substring(0, s.length()).equals(s));
		} else {
			return false;
		}
	}

	public void parseParameters(String[] args) {
		for (int i = 0; i < args.length; i++) {
			String param = (String) args[i];

			try {
				if (param.startsWith("-")) {
					if (param.equals("-batch")) {
						setBatchMode(true);
					} else if (param.equals("-compact")) {
						setCompactMode(true);
					} else if (param.equals("-warnundeclaredchemicals")) {
						setWarnUndeclaredChemicals();
					} else if (param.equals("-nowarnundeclaredchemicals")) {
						setNoWarnUndeclaredChemicals();
					} else if (prefixEquals(param, "-sim:")) {
						setSimFile(param.substring(5));
					} else if (prefixEquals(param, "-system:")) {
						setSystemModel(param.substring(8));
					} else if (prefixEquals(param, "-out:")) {
						setOutFile(param.substring(5));
					} else if (prefixEquals(param, "-err:")) {
						setErrFile(param.substring(5));
					} else if (prefixEquals(param, "-2d") || (prefixEquals(param, "-2D"))) {
						isDisplay3D = false;
					} else if (prefixEquals(param, "-3d") || (prefixEquals(param, "-3D"))) {
						isDisplay3D = true;
					} else if (prefixEquals(param, "-dna:")) {
						int dnaValue =
							new Integer(param.substring(5)).intValue();
						setDNAinit(dnaValue);
					} else if (prefixEquals(param, "-stat:")) {
						setStatsFile(param.substring(6));
					} else if (prefixEquals(param, "-seed:")) {
						setRandomSeed(new Long(param.substring(6)).longValue());
					} else if (prefixEquals(param, "-logall")) {
						setLogAll();
					} else if (prefixEquals(param, "-logcounts")) {
						setLogCounts();
					} else if (prefixEquals(param, "-dither:")) {
						setDither(new Double(param.substring(8)).doubleValue());
					} else if (prefixEquals(param, "-noscale")) {
						setNoScale();
					} else if (prefixEquals(param, "-report:")) {
						int level = new Integer(param.substring(8)).intValue();
						util.Report.setReportLevel(level);
					} else if (prefixEquals(param, "-csvstats")) {
						setCSVStats(true);
					} else if (prefixEquals(param, "-steps:")) {
						int numSteps =
							new Integer(param.substring(7)).intValue();
						setMaxSteps(numSteps);
					} else if (prefixEquals(param, "-program:")) {
						addProgramFile(param.substring(9));
					} else {
						System.err.println("Unrecognized option: " + param);
					}
				} else {
					// Its a file name
					int extchar = param.indexOf('.');

					if (extchar == -1) {
						System.err.println("Unrecognized file: " + param);
					} else {
						String extension = param.substring(extchar);

						if (extension.equals(".cpa")) {
							addProgramFile(param);
						} else if (extension.equals(".sim")) {
							setSimFile(param);
						} else {
							System.err.println(
								"Unrecognized file extension: " + extension);
						}
					}
				}
			} catch (StringIndexOutOfBoundsException e) {
				System.err.println(e.toString());
				System.err.println(param);
			}
		}
	}

	public boolean isDisplay3D() {
		return isDisplay3D;
	}
}
