package simulator;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.*;
import parsers.CPA.CellParameters;
import util.*;

import sysmodels.ChemicalDiffusion;
import sysmodels.Environment;
import sysmodels.EVector;

public abstract class SystemModel {
	private Simulation sim;
	protected StatsTable statistics;
	protected int numCellsKilled;
	protected int numCellsDied;
	protected long lastStepTime = 0;
	protected long totalTime = 0;
	private Environment env;
	private double concern = 0.0; //-------------------

	public abstract void cellSuicide(Cell c);
	public abstract int numberOfCells();
	public abstract void killCell(Cell c);
	protected abstract Cell getNewCell(Vector states, int programNum);
	protected abstract void installNewCell(
		Vector states,
		Position pos,
		int programNum);
	protected abstract boolean displaceCells(Cell cell1, Cell cell2);
	public abstract Enumeration adjacentCells(Cell c, Direction dir);

	//this was hiding in the middle of the file
	//protected CellProgram cellProgram;
	protected CellProgram cellProgram;

	public SystemModel(Simulation p_sim) {
		numCellsKilled = 0;
		numCellsDied = 0;
		sim = p_sim;
		env = new Environment();

	}

	protected void Report(String s) {
		Report.message("[StandardModel] " + s);
	}

	public Simulation getSimulation() {
		return sim;
	}

	public void init(Vector args) {
		cellProgram = getSimulation().getCellProgram();
		statistics = new StatsTable();
		lastStepTime = 0;
		totalTime = 0;
		//env = new Environment();
	}

	public void generateOutput(BufferedWriter outputWriter)
		throws IOException {
		Enumeration enum = allCells();
		while (enum.hasMoreElements()) {
			Cell cd = (Cell) enum.nextElement();
			String s =
				cd.getCurrentProgramState()
					+ " "
					+ cd.getX()
					+ " "
					+ cd.getY()
					+ " "
					+ cd.getZ();
			//-----------

			if (outputWriter == null)
				getSimulation().writeOutputLine(s);
			else {
				outputWriter.write(s, 0, s.length());
				outputWriter.newLine();
			}
		}
	}

	protected void runInstruction(Vector instruction) {
		int index = 0;
		if (instruction.size() == 0) {
			return; // Empty line
		}

		String instrType = (String) instruction.elementAt(index++);

		if (instrType.equals("new")) {
			String dummy;
			Vector states = new Vector();
			String[] statesArr;
			int numStates = 0;

			dummy = (String) instruction.elementAt(index++);
			if (dummy.compareTo("(") != 0) {
				states.add(dummy);
				numStates = 1;
			} else {
				dummy = (String) instruction.elementAt(index++);
				while (dummy.compareTo(")") != 0) {
					states.add(dummy);
					numStates++;
					dummy = (String) instruction.elementAt(index++);
				}
			}

			//statesArr = states.toArray();
			//statesArr = new String[states.size()];
			//states.toArray(statesArr);

			double x = ((Double) instruction.elementAt(index++)).doubleValue();
			double y = ((Double) instruction.elementAt(index++)).doubleValue();
			double z = ((Double) instruction.elementAt(index++)).doubleValue();

			installNewCell(states, new Position(x, y, z), 0);
		} else if (instrType.equals("populate")) {
			// populate <state> x1 y1 z1 x2 y2 z2 inc
			// populate [normal] <state> x1 y1 z1 x2 y2 z2 numcells
			// populate [normal] [circle] <state> x1 y1 radius numcells

			// fill up the region (x1, y1, z1) -> (x2, y2, z2) with cells in state
			// with distance inc between cells.

			String state = (String) instruction.elementAt(index++);
			boolean normal = false;
			boolean circle = false;
			boolean oval = false;
			double ovalspread = 1.0;
			
			if (state.equals("[normal]")) {
				normal = true;
				state = (String) instruction.elementAt(index++);
			}

			if (state.equals("[circle]")) {
				circle = true;
				state = (String) instruction.elementAt(index++);
			}

			if (state.equals("[oval]")) {
				oval = true;
				state = (String) instruction.elementAt(index++);
			}
			
			Vector states = new Vector ();
			states.add (state);

			double x1 = ((Double) instruction.elementAt(index++)).doubleValue();
			double y1 = ((Double) instruction.elementAt(index++)).doubleValue();
			
			if (circle || oval) {
				double radius =
					((Double) instruction.elementAt(index++)).doubleValue();
				if (oval) {
					ovalspread = ((Double) instruction.elementAt(index++)).doubleValue();	
				}

				int numcells =
					(int) ((Double) instruction.elementAt(index++))
						.doubleValue();

				// Randomly populate the circle by choosing distance [0, radius> and direction [0, 360>
				for (int i = 0; i < numcells; i++) {
					double dist = (SimRandom.nextDouble() * radius);
					double dir = (SimRandom.nextDouble() * 2 * Math.PI);

					double x = Math.cos(dir) * dist;
					double y = Math.sin(dir) * dist;

					if (oval) {
						x *= ovalspread;
					}
					
					installNewCell(states, new Position(x1 + x, y1 + y, 0.0), 0);
				}

			} else {

				double z1 =
					((Double) instruction.elementAt(index++)).doubleValue();
				double x2 =
					((Double) instruction.elementAt(index++)).doubleValue();
				double y2 =
					((Double) instruction.elementAt(index++)).doubleValue();
				double z2 =
					((Double) instruction.elementAt(index++)).doubleValue();

				if (normal) {
					int numcells =
						(int) ((Double) instruction.elementAt(index++))
							.doubleValue();

					for (int i = 0; i < numcells; i++) {
						double posx = x1 + (SimRandom.nextDouble() * (x2 - x1));
						double posy = y1 + (SimRandom.nextDouble() * (y2 - y1));
						double posz = z1 + (SimRandom.nextDouble() * (z2 - z1));

						installNewCell(states, new Position(posx, posy, posz), 0);
					}
				} else {
					double inc =
						((Double) instruction.elementAt(index++)).doubleValue();

					Assert.check(inc > 0.0);
					for (double z = z1; z <= z2; z += inc) {
						for (double y = y1; y <= y2; y += inc) {
							for (double x = x1; x <= x2; x += inc) {
								installNewCell(states, new Position(x, y, z), 0);
							}
						}
					}
				}
			}
		} else if (instrType.equals("force")) {
			double x1, y1, z1;
			double x2, y2, z2;
			x1 = ((Double) instruction.elementAt(index++)).doubleValue();
			y1 = ((Double) instruction.elementAt(index++)).doubleValue();
			z1 = ((Double) instruction.elementAt(index++)).doubleValue();

			x2 = ((Double) instruction.elementAt(index++)).doubleValue();
			y2 = ((Double) instruction.elementAt(index++)).doubleValue();
			z2 = ((Double) instruction.elementAt(index++)).doubleValue();

			env.setForceVector(
				new EVector(
					new Position(x1, y1, z1),
					new Position(x2, y2, z2)));
		}else if (instrType.equals("segmentF")) {
		String pole = (String) instruction.elementAt(index++);
		double pos1 = 0.0;
		//double pos2 = 0.0;
		double pos = 0.0;
		Cell cell1 = null;
		//Cell cell2 = null;
			
		Enumeration enums = allCells();
		while (enums.hasMoreElements()) {
			Cell cell = (Cell) enums.nextElement();
			if (pole.equals("x")) {
				pos = cell.getPosition().getX();
			} else if (pole.equals("y")) {
				pos = cell.getPosition().getY();
			} else if (pole.equals("z")) {
				pos = cell.getPosition().getZ();
			}

			if (pos < pos1) {
				pos1 = pos;
				cell1 = cell;
			}/* else if (pos < pos2) {
				pos2 = pos;
				cell2 = cell;					
			}*/
		}
			
		if (cell1 == null) {
		Report.message (0, "Didn't find cell 1(anterior)!");
		} else {
					cell1.setCurrentProgramState("R");
					Report.message (0, "Setting state: " + cell1);
		}
			
		/*if (cell2 == null) {
			Report.message (0, "Didn't find cell 2!");
		} else {
			
		cell2.setExpressedState("L");
		Report.message (0, "Setting state: " + cell2);
		} 
		*/
		
		}else if (instrType.equals("segmentF")) {
		String pole = (String) instruction.elementAt(index++);
		double pos1 = 0.0;
		//double pos2 = 0.0;
		double pos = 0.0;
		Cell cell1 = null;
		//Cell cell2 = null;
			
		Enumeration enums = allCells();
		while (enums.hasMoreElements()) {
			Cell cell = (Cell) enums.nextElement();
			if (pole.equals("x")) {
				pos = cell.getPosition().getX();
			} else if (pole.equals("y")) {
				pos = cell.getPosition().getY();
			} else if (pole.equals("z")) {
				pos = cell.getPosition().getZ();
			}

			if (pos < pos1) {
				pos1 = pos;
				cell1 = cell;
			}/* else if (pos < pos2) {
				pos2 = pos;
				cell2 = cell;					
			}*/
		}
			
		if (cell1 == null) {
		Report.message (0, "Didn't find cell 1(anterior)!");
		} else {
					cell1.setCurrentProgramState("R");
					Report.message (0, "Setting state: " + cell1);
		}
			
		/*if (cell2 == null) {
			Report.message (0, "Didn't find cell 2!");
		} else {
			
		cell2.setExpressedState("L");
		Report.message (0, "Setting state: " + cell2);
		} 
		*/
		
		} else if (instrType.equals("segment")) {
			String pole = (String) instruction.elementAt(index++);
			double pos1 = 0.0;
			double pos2 = 0.0;
			double pos = 0.0;
			Cell cell1 = null;
			Cell cell2 = null;

			Enumeration enums = allCells();
			while (enums.hasMoreElements()) {
				Cell cell = (Cell) enums.nextElement();
				if (pole.equals("x")) {
					pos = cell.getPosition().getX();
				} else if (pole.equals("y")) {
					pos = cell.getPosition().getY();
				} else if (pole.equals("z")) {
					pos = cell.getPosition().getZ();
				}

				if (pos > pos1) {
					pos1 = pos;
					cell1 = cell;
				} else if (pos < pos2) {
					pos2 = pos;
					cell2 = cell;
				}
			}

			if (cell1 == null) {
				Report.message(0, "Didn't find cell 1!");
			} else {
				cell1.setCurrentProgramState("R");
				Report.message(0, "Setting state: " + cell1);
			}

			if (cell2 == null) {
				Report.message(0, "Didn't find cell 2!");
			} else {

				cell2.setCurrentProgramState("L");
				Report.message(0, "Setting state: " + cell2);
			}
			/*
						String chemical1 = (String) instruction.elementAt(index++);
						double chem1Conc =
							((Double) instruction.elementAt(index++)).doubleValue();
						double chem1Rate =
							((Double) instruction.elementAt(index++)).doubleValue();
						String chemical2 = (String) instruction.elementAt(index++);
						double chem2Conc =
							((Double) instruction.elementAt(index++)).doubleValue();
						double chem2Rate =
							((Double) instruction.elementAt(index++)).doubleValue();
						long t = System.currentTimeMillis();
						
						ChemicalDiffusion cd1 =
							new ChemicalDiffusion(
								c1.getPosition(),
								chemical1,
								chem1Conc,
								chem1Rate,
								t);
			
						
						ChemicalDiffusion cd2 =
							new ChemicalDiffusion(
								c2.getPosition(),
								chemical2,
								chem2Conc,
								chem2Rate,
								t);
						try {
							env.add(cd1);
							env.add(cd2);
						} catch (NullPointerException e) {
							System.err.println("unable to add environment vector");
						}
						System.out.println("\nchem "+chemical1+"conc "+chem1Conc);
								*/

		} else if (instrType.equals("kill")) {
			double x1, y1, z1, x2, y2, z2, radius;
			String area = (String) instruction.elementAt(index++);
			if (area.equals("sphere")) {
				x1 = ((Double) instruction.elementAt(index++)).doubleValue();
				y1 = ((Double) instruction.elementAt(index++)).doubleValue();
				z1 = ((Double) instruction.elementAt(index++)).doubleValue();
				radius =
					((Double) instruction.elementAt(index++)).doubleValue();
				System.err.println(
					"Killing sphere: "
						+ x1
						+ ", "
						+ y1
						+ ", "
						+ z1
						+ " radius: "
						+ radius);
				killSphere(new Position(x1, y1, z1), radius);
			} else if (area.equals("block")) {
				x1 = ((Double) instruction.elementAt(index++)).doubleValue();
				y1 = ((Double) instruction.elementAt(index++)).doubleValue();
				z1 = ((Double) instruction.elementAt(index++)).doubleValue();
				x2 = ((Double) instruction.elementAt(index++)).doubleValue();
				y2 = ((Double) instruction.elementAt(index++)).doubleValue();
				z2 = ((Double) instruction.elementAt(index++)).doubleValue();
				killBlock(x1, y1, z1, x2, y2, z2);
			} else if (area.equals("random")) {
				String state = null;

				if (instruction.elementAt(index) instanceof String) {
					state = (String) instruction.elementAt(index++);
				}

				double fraction =
					((Double) instruction.elementAt(index++)).doubleValue();
				if (fraction < 0.0 || fraction > 1.0) {
					System.err.println(
						"Malformed kill instruction: kill random "
							+ fraction
							+ ": the kill fraction should be between 0 and 1");
					return;
				}

				if (state != null) {
					killRandomly(fraction, state);
				} else {
					killRandomly(fraction);
				}
			} else {
				System.err.println("Malformed kill instruction: kill " + area);
			}
		} else if (instrType.equals("Output")) {
			String s = null, file = null;
			try {
				file = (String) instruction.elementAt(index++);
			} catch (ArrayIndexOutOfBoundsException e) {
				System.err.println(e.toString());
			}
			System.out.println("Here: " + file);
			BufferedWriter outputWriter = null;
			try {
				if (file != null) {
					outputWriter = new BufferedWriter(new FileWriter(file));
				}

				generateOutput(outputWriter);

				if (outputWriter != null)
					outputWriter.close();
			} catch (IOException e) {
				System.err.println(e.toString());
			}
		} else if (instrType.equals("Statistics")) {
			try {
				Enumeration enum = allCells();
				while (enum.hasMoreElements()) {
					Cell cd = (Cell) enum.nextElement();
					getSimulation().writeOutputLine(
						cd.getX()
							+ " "
							+ cd.getY()
							+ " "
							+ cd.getZ()
							+ " "
							+ cd.getColor().getR()
							+ " ");
				}
			} catch (IOException e) {
			}
		}
	}

	/**
	 * @param fraction
	 * @param state
	 */
	public void killRandomly(double fraction, String state) {
		Vector killCells = new Vector();
		for (Enumeration cenum = allCells(); cenum.hasMoreElements();) {
			Cell cell = (Cell) cenum.nextElement();
			double rand = SimRandom.nextDouble();
			if (rand <= fraction) {
				if (state == null || cell.matchesState(state)) {
					killCells.add(cell);
				}
			}
		}

		for (Enumeration kenum = killCells.elements();
			kenum.hasMoreElements();
			) {
			killCell((Cell) kenum.nextElement());
		}
	}

	protected void updateStatistics() {
		statistics.setValue("Simulation Step", getSimulation().getStep());
		statistics.setValue("Number of cells", numberOfCells());
		statistics.setValue("Last Step Time", lastStepTime);
		statistics.setValue("Total Time", totalTime);
		statistics.setValue("Total Memory", Runtime.getRuntime().totalMemory());
		statistics.setValue("Number of cells killed", numCellsKilled);
		statistics.setValue("Number of cells self-died", numCellsDied);
	}

	public void emitChemicals(Cell c, Vector emissions) {
		Report("Emissions: " + emissions + " from " + c);
		for (Enumeration emits = emissions.elements();
			emits.hasMoreElements();
			) {
			Enumeration cells;
			CellEmission emit = (CellEmission) emits.nextElement();
			Direction dir = emit.getDirection();

			for (Enumeration ce = adjacentCells(c, dir);
				ce.hasMoreElements();
				) {
				Cell neighbor = (Cell) ce.nextElement();
				Report("Sending emission " + emit + " to " + neighbor);
				neighbor.receiveChemical(c, emit);
			}
		}
	}

	public void diffuseChemicals(Cell c, Vector diffusions) {
		Report("Diffusions: " + diffusions);
		diffuseChemicals(c.getPosition(), diffusions);
	}

	public double getConcentration(Position p, String chem, int step) {
		double conc;
		conc = env.findConcentration(chem, p, step);
		concern = conc;
		return conc;
	}

	private void diffuseChemicals(Position p, Vector diffusions) {

		Enumeration diffusionEnum = diffusions.elements();

		while (diffusionEnum.hasMoreElements()) {
			CellDiffusion diffusion =
				(CellDiffusion) diffusionEnum.nextElement();
			int step = getSimulation().getStep();
			ChemicalDiffusion cd =
				new ChemicalDiffusion(
					p,
					diffusion.getChemical(),
					diffusion.getConcentration(),
					diffusion.getDiffuseRate(),
					diffusion.getSpreadRate(),
					diffusion.getEvaporateRate(),
					diffusion.getDiffusionModel(),
					step);

			try {
				env.add(cd);
			} catch (NullPointerException e) {
				System.err.println("unable to add environment vector");
			}
		}
	}

	public StatsTable getStatistics() {
		return statistics;
	}

	// Enumeration that yields all cells in the system
	public abstract Enumeration allCells();

	// Carry out emissions and diffusions based on the current state of the cell.
	private void executeStateSecretions() {
		Enumeration enum = allCells();

		while (enum.hasMoreElements()) {
			Cell cell = (Cell) enum.nextElement();
			cell.executeStateSecretions();
		}
	}

	public void runCellProgramForOneStep() {
		long timeBefore = System.currentTimeMillis();

		// Before running the new step, all emissions from the previous step must
		// be sent to neighboring cells and all new cells become old.

		Report("--- Running one step ---");

		Enumeration enum = allCells();
		while (enum.hasMoreElements()) {
			Cell cell = (Cell) enum.nextElement();
			if (cell.isNewCell()) {
				cell.setNewCell(false);
			} else {
				cell.disposeSensedChemicals();
			}
		}

		//env.clearDiffusions();
		Report("--- Executing Secretions ---");
		executeStateSecretions();

		// Cannot use enum here, divisions create new cells
		enum = allCells();

		Report("--- Running Cell Programs ---");
		while (enum.hasMoreElements()) {
			Cell cell = (Cell) enum.nextElement();
			Assert.check(!cell.isNewCell());
			cell.runCellProgramForOneStep();
		}

		long timeAfter = System.currentTimeMillis();
		lastStepTime = timeAfter - timeBefore;
		totalTime += lastStepTime;
		updateStatistics();
		Report("--- Step Complete ---");
	}
	
	protected Position newCellPosition(Position pos, Direction dir) {
		return pos.offset(dir);
	}

	public void performCellDivision(
		Cell cell,
		Bindings bindings,
		CellParameters params1,
		CellParameters params2,
		Direction divisionDir) {
		Report(
			"Performing division: "
				+ divisionDir
				+ " ==> "
				+ params1
				+ " / "
				+ params2);

		//create a copy of the states array and then put the new
		//state in the correct position
		Vector cell2States = new Vector();
		for (int program = 0; program < cell.getStates().length; program++) {
			cell2States.add(cell.getStates()[program]);
		}
		cell2States.set(cell.getCurrentProgramNum(), params2.getState());
		//create a new cell with this state array
		Cell cell2 = getNewCell(cell2States, cell.getCurrentProgramNum());
		cell2.reflectParameters(cell, bindings, params2);

		if (cell2 == null) {
			Assert.badBranch("Not enough Memory");
		}

		if (divisionDir == null) {
			Assert.badBranch(
				"Direction for division null: " + params1 + " / " + params2);
		}

		Position pstart = cell.getPosition();
		Position pnew = newCellPosition(pstart, divisionDir);
		cell2.setPosition(pnew);

		if (displaceCells(cell, cell2)) {
			// cell2 is added, change cell1 state
			cell.setCurrentProgramState(params1.getState());
			//! really should create a new cell!
			cell.reflectParameters(cell, bindings, params1);
		} else {
			// No space for cell2
			killCell(cell2);
		}
	}

	protected void killBlock(
		double x1,
		double y1,
		double z1,
		double x2,
		double y2,
		double z2) {
		Vector killCells = new Vector();
		Position p1 = new Position(x1, y1, z1);
		Position p2 = new Position(x2, y2, z2);

		for (Enumeration cenum = allCells(); cenum.hasMoreElements();) {
			Cell cell = (Cell) cenum.nextElement();
			if (cell.getPosition().withinBlock(p1, p2)) {
				killCells.add(cell);
			}
		}

		for (Enumeration kenum = killCells.elements();
			kenum.hasMoreElements();
			) {
			killCell((Cell) kenum.nextElement());
		}
	}

	/**
	 * @param position
	 * @param radius
	 */
	protected void killSphere(Position position, double radius) {
		Vector killCells = new Vector();

		for (Enumeration ce = allCells(); ce.hasMoreElements();) {
			Cell cell = (Cell) ce.nextElement();
			if (cell.getPosition().distance(position) <= radius) {
				// System.err.println ("Killing: " + cell + " at " + cell.getPosition ());
				killCells.add(cell);
			} else {
				// System.err.println ("Not killing: " + cell + " at " + cell.getPosition ());
			}
		}

		for (Enumeration kenum = killCells.elements();
			kenum.hasMoreElements();
			) {
			killCell((Cell) kenum.nextElement());
		}

	}

	public void killRandomly(double fraction) {
		killRandomly(fraction, null);
	}

	public double getDither() {
		return sim.getDither();
	}
}
