/*
 * Created on Jun 14, 2003
 */
package sysmodels.floating;
import simulator.*;
import java.util.Vector;
import java.util.Enumeration;
import util.*;

/**
 * @author David Evans
 */

public class FloatingModel extends SystemModel {
	// In the FloatingModel, cells have floating point locations.
	// The FloatingModel must keep track of adjacent cells, and
	// adjust locations when cells divide.
	
	protected Vector cells;
	// Location map: for each unit square in space, keep a bucket of the cells
	// in that square.  To find adjacent cells, look in all adjacent buckets for
	// the closest cell in each direction.

	public FloatingModel(Simulation p_sim) {
		super(p_sim);
		cells = new Vector ();
		Report.message (0, "Creating Floating Model");
	}
	
	public int numberOfCells () {
		return cells.size();
	}
	
	public Enumeration allCells () {
		return new Vector (cells).elements ();
	}
	
	// effects Returns a Vector containing all cells within distance of p.
	private Vector cellsWithinDistance (Position p, double distance) {
		Vector res = new Vector ();
		
		for (Enumeration enum = cells.elements (); enum.hasMoreElements (); ) {
			Cell c = (Cell) enum.nextElement ();
			
			if (c.getPosition().distance (p) <= distance) {
				res.add (c);
			}
		}
		
		return res;		
	}
	
	public Enumeration adjacentCells(Cell c, Direction dir) {
		// We approximate adjacentCell using all cells within distance 1.0
		Vector res = cellsWithinDistance (c.getPosition (), 1.0);
		Assert.check (res.remove(c)); 
		return res.elements ();
	}
	
	private void killCell(Cell cell, boolean self) {
		if (self) {
			numCellsDied++;
		} else {
			numCellsKilled++;
		}
		
		Assert.check (cells.remove(cell));
	}
		
	public void killCell(Cell cd) {
		killCell (cd, false);
	}

	public void cellSuicide (Cell cell) {
		   killCell(cell, true);
	}  

	public boolean displaceCells (Cell cell1, Cell cell2) {
		/*
		 * When a new cell is added, if a cell is already too close to
		 * it, we need to move them.
		 */
		// if (cell1.getPosition().distance (cell2.getPosition ())
		
		return true;   
	}

	public Cell getNewCell (Vector states, int programNum) {
		Cell cell =
			getSimulation().createCell (cells.size (), this, cellProgram, states, programNum);
		cells.add(cell);
		cell.setNewCell(true);
		Report ("Created new cell: " + cell);
		return cell;	
	}
	
	public void installNewCell (Vector states, Position pos, int programNum) {
		Cell newcell = getNewCell (states, programNum);
		newcell.setPosition (pos);
		cells.add (newcell);			
	}
	
	protected Position newCellPosition (Position pos, Direction dir) {
		Position pnew = pos.offset (dir);
		double scale = 0.25; 
		pnew.setX (pnew.getX () + (scale / 2.0) - (scale * SimRandom.nextDouble ()));
		pnew.setY (pnew.getY () + (scale / 2.0) - (scale * SimRandom.nextDouble ()));
		pnew.setZ (pnew.getZ () + (scale / 2.0) - (scale * SimRandom.nextDouble ()));
		return pnew;
	}

	static class FloatingModelFactory extends SystemFactory {
		public SystemModel create(Simulation p_sim, String modelOptions) {
			FloatingModel sb = new FloatingModel(p_sim);
			Vector vec = new Vector();
			vec.add(modelOptions);
			sb.init(vec);
			return sb;
		}
	}
	
	static {
		Simulation.sysModels.put(
			new String("sysmodels.floating.FloatingModel"),
			new FloatingModelFactory());
	}
}
