package sysmodels.standard;

import java.util.*;
import simulator.*;
import util.*;

public class StandardModel extends SystemModel {
	static class StandardModelFactory extends SystemFactory {
		public SystemModel create(Simulation p_sim, String modelOptions) {
			StandardModel sb = new StandardModel(p_sim);
			Vector vec = new Vector();
			vec.add(modelOptions);
			sb.init(vec);
			return sb;
		}
	}

	static {
		Simulation.sysModels.put(
			new String("sysmodels.standard.StandardModel"),
			new StandardModelFactory());
	}

	final static private int MAXDEPTH = 100;
	final static private int offsetX = 50;
	final static private int offsetY = 50;
	final static private int offsetZ = 50;
	// Size of each dimension (max MAXDEPTH^3 cells)
	protected Cell cells[][][]; // Cells are stored in a 3D matrix 
	protected Vector IDMappings;

	public StandardModel(Simulation p_sim) {
		super(p_sim);
	}

	public Enumeration adjacentCells(Cell c, Direction d) {
		Vector res = new Vector();
		double x = c.getX();
		double y = c.getY();
		double z = c.getZ();

		if (d == null || (d.isAll() || d.equals(Direction.XP))
			&& locationInRange(x + 1, y, z)) {
			Cell ce = getCell(x + 1, y, z);

			if (ce != null) {
				res.add(ce);
			}
		}

		if (d == null || (d.isAll() || d.equals(Direction.XM))
			&& locationInRange(x - 1, y, z)) {
			Cell ce = getCell(x - 1, y, z);

			if (ce != null && ce != c) {
				res.add(ce);
			}
		}

		if (d == null || (d.isAll() || d.equals(Direction.YP))
			&& locationInRange(x, y + 1, z)) {
			Cell ce = getCell(x, y + 1, z);

			if (ce != null && ce != c) {
				res.add(ce);
			}
		}

		if (d == null || (d.isAll() || d.equals(Direction.YM))
			&& locationInRange(x, y - 1, z)) {
			Cell ce = getCell(x, y - 1, z);

			if (ce != null && ce != c) {
				res.add(ce);
			}
		}

		if (d == null || (d.isAll() || d.equals(Direction.ZP))
			&& locationInRange(x, y, z + 1)) {
			Cell ce = getCell(x, y, z + 1);

			if (ce != null && ce != c) {
				res.add(ce);
			}
		}

		if (d == null || (d.isAll() || d.equals(Direction.ZM))
			&& locationInRange(x, y, z - 1)) {
			Cell ce = getCell(x, y, z - 1);

			if (ce != null && ce != c) {
				res.add(ce);
			}
		}

		Assert.check(res.size() <= 6);
		return res.elements();
	}

	public void cellSuicide(Cell cell) {
		selfKillCell(cell);
	}

	// StandardModel does not allow any displacement
	protected boolean displaceCells(Cell cell1, Cell cell2) {
		double dx = cell2.getX() - cell1.getX();
		double dy = cell2.getY() - cell1.getY();
		double dz = cell2.getZ() - cell1.getZ();
		Cell cell = cell2;

		double x = cell.getX();
		double y = cell.getY();
		double z = cell.getZ();

		if (!locationInRange(x, y, z)) {
			getSimulation().writeErrorLine(
				"Cell location out of range: " + x + ", " + y + ", " + z);
			return false;
		} else {
			Cell cdOther = getCell(x, y, z);
			if (cdOther == null) {
				setCell((int) x, (int) y, (int) z, cell);
				return true;
			} else {
				// kill the other cell, and replace it with this cell
				killCell(cdOther);
				setCell((int) x, (int) y, (int) z, cell);
				return true;
				/* was:
				return false;
				*/
			}
		}
	}

	private Cell getCell(int id) {
		Enumeration enum = IDMappings.elements();

		while (enum.hasMoreElements()) {
			Cell cell = (Cell) enum.nextElement();

			if (id == cell.getID())
				return cell;
		}
		return null;
	}

	protected Cell getCell(double x, double y, double z) {
		int locx = (int) x;
		int locy = (int) y;
		int locz = (int) z;
		Assert.check(locationInRange(locx, locy, locz));
		return cells[locx + offsetX][locy + offsetY][locz + offsetZ];
	}

	protected Cell getNewCell(Vector states, int programNum) {
		Cell cell =
			getSimulation().createCell(
				IDMappings.size(),
				this,
				cellProgram,
				states,
				programNum);
		IDMappings.add(cell);
		cell.setNewCell(true);
		Report("Created new cell: " + cell);
		return cell;
	}

	protected void installNewCell(Vector states, Position pos, int programNum) {
		Cell cell = getNewCell(states, programNum);
		cell.setPosition(pos);
		setCell((int) pos.getX(), (int) pos.getY(), (int) pos.getZ(), cell);
	}

	public void init(Vector args) {
		super.init(args);
		IDMappings = new Vector();
		cells = new Cell[MAXDEPTH][MAXDEPTH][MAXDEPTH];
	}

	public void killCell(Cell cd) {
		killCell(cd, false);
	}

	private void killCell(Cell cd, boolean self) {
		int x = (int) cd.getX();
		int y = (int) cd.getY();
		int z = (int) cd.getZ();

		if (self) {
			numCellsDied++;
		} else {
			numCellsKilled++;
		}

		Assert.check(IDMappings.remove(cd));

		if (getCell(x, y, z) == cd) {
			setCell(x, y, z, null);
		} else {
			Report("kill mismatch: " + cd + " / " + getCell(x, y, z));
		}
	}

	protected boolean locationInRange(int x, int y, int z) {
		return (
			(x + offsetX >= 0 && x + offsetX < MAXDEPTH)
				&& (y + offsetY >= 0 && y + offsetY < MAXDEPTH)
				&& (z + offsetZ >= 0 && z + offsetZ < MAXDEPTH));
	}

	protected boolean locationInRange(double x, double y, double z) {
		return locationInRange((int) x, (int) y, (int) z);
	}

	public int numberOfCells() {
		return IDMappings.size();
	}

	public Enumeration allCells() {
		return ((Vector) IDMappings.clone()).elements();
	}

	private void selfKillCell(Cell cd) {
		killCell(cd, true);
	}

	protected void setCell(int x, int y, int z, Cell cell) {
		Assert.check(locationInRange(x, y, z));
		cells[x + offsetX][y + offsetY][z + offsetZ] = cell;
	}

	private void storeChemicals(Cell cell, Vector emissions) {
		if (emissions != null) {
			Enumeration emissionEnum = emissions.elements();

			while (emissionEnum.hasMoreElements()) {
				CellEmission emission =
					(CellEmission) emissionEnum.nextElement();
				cell.addSensedChemical(emission);
			}
		}
	}
}
