package util;

public class Direction {
	//
	// Direction is immutable.
	// 
	// A direction is a represented by a 2 bits for each dimension
	// Using an 32 bit int, we have enough bits for 16 dimensions (but probably
	// only need 3...)
	//
	// The value for each dimension is:
	//     1   Up/North/Right
	//     0
	//    -1   Down/South/Left
	//

	private char rep[];

	static public final Direction UNDEFINED = new Direction("UN");
	static public final Direction ALL = new Direction("AL");
	static public final Direction ANY = new Direction("AY");
	static public final Direction OPPOSITE = new Direction("OP");
	static public final Direction SAME = new Direction("SA");
	static public final Direction PERPENDICULAR = new Direction("PU");

	static public final Direction XM = new Direction('X', '-');
	static public final Direction XP = new Direction('X', '+');
	static public final Direction YM = new Direction('Y', '-');
	static public final Direction YP = new Direction('Y', '+');
	static public final Direction ZM = new Direction('Z', '-');
	static public final Direction ZP = new Direction('Z', '+');

	public Direction(String dir) {
		Assert.check(dir.length() == 2);
		rep = new char[2];
		rep[0] = dir.charAt(0);
		rep[1] = dir.charAt(1);
	}

	public Direction(char d1, char d2) {
		rep = new char[2];
		rep[0] = d1;
		rep[1] = d2;
	}

	// Turns a direction into a position vector
	public Position canonicalize() {
		double x = getX();
		double y = getY();
		double z = getZ();

		return new Position(x, y, z);
	}

	// bleech!  There's got to be a better way to do this...
	private static double roundClose(double d) {
		if (d < 0.0) {
			double neg = -d;
			if (Math.abs(neg - Math.floor(neg)) < 0.01) {
				return (double) - Math.floor(neg);
			} else {
				return d;
			}
		} else {
			if (Math.abs(d - Math.floor(d)) < 0.01) {
				return (double) Math.floor(d);
			} else {
				return d;
			}
		}
	}

	public static Direction fromPosition(Position p) {
		double x = roundClose(p.getX());
		double y = roundClose(p.getY());
		double z = roundClose(p.getZ());
		Direction res;

		if (x == 1.0 && y == 0.0 && z == 0.0) {
			res = Direction.XP;
		} else if (x == -1.0 && y == 0.0 && z == 0.0) {
			res = Direction.XM;
		} else if (x == 0.0 && y == 1.0 && z == 0.0) {
			res = Direction.YP;
		} else if (x == 0.0 && y == -1.0 && z == 0.0) {
			res = Direction.YM;
		} else if (x == 0.0 && y == 0.0 && z == 1.0) {
			res = Direction.ZP;
		} else if (x == 0.0 && y == 0.0 && z == -1.0) {
			res = Direction.ZM;
		} else {
			Report.fatalError(
				"Position is not a direction: "
					+ p
					+ " ("
					+ x
					+ ", "
					+ y
					+ ", "
					+ z);
			res = Direction.XP;
		}

		return res;
	}

	public boolean equals(Object obj) {
		if (obj instanceof Direction) {
			Direction other = (Direction) obj;
			return rep[0] == other.rep[0] && rep[1] == other.rep[1];
		} else {
			return false;
		}
	}

	static public Direction indexDirection(int dir) {
		Direction res;

		switch (dir) {
			case 0 :
				res = XM;
				break;
			case 1 :
				res = XP;
				break;
			case 2 :
				res = YM;
				break;
			case 3 :
				res = YP;
				break;
			case 4 :
				res = ZM;
				break;
			case 5 :
				res = ZP;
				break;
			default :
				Assert.badBranch("Bad direction: " + dir);
				res = UNDEFINED;
		}

		Assert.check(res.directionIndex() == dir);
		return res;
	}

	public boolean isSpecialDirection() {
		return (
			this == ALL
				|| this == ANY
				|| this == UNDEFINED
				|| this == OPPOSITE
				|| this == SAME
				|| this == PERPENDICULAR);
	}

	public boolean isUndefined() {
		return this == UNDEFINED;
	}

	public boolean isSame() {
		return this == SAME;
	}

	public boolean isOpposite() {
		return this == OPPOSITE;
	}

	public boolean isAny() {
		return this == ANY;
	}

	public boolean isAll() {
		return this == ALL;
	}

	public boolean isPerpendicular() {
		return this == PERPENDICULAR;
	}

	static public int numDirections(int dimensions) {
		return dimensions * 2;
	}

	public int directionIndex() {
		Assert.check(!isSpecialDirection());
		int index = 0;

		switch (rep[0]) {
			case 'X' :
				index = 0;
				break;
			case 'Y' :
				index = 2;
				break;
			case 'Z' :
				index = 4;
				break;
			default :
				Assert.badBranch("Strange direction: " + rep[0]);
		}

		if (rep[1] == '+') {
			index++;
		}

		return index;
	}

	private static int dirValue(char dir) {
		if (dir == '-') {
			return -1;
		}
		if (dir == '+') {
			return 1;
		}
		Assert.badBranch("Bad dir: " + dir);
		return 0;
	}

	public int getX() {
		Assert.check(!isSpecialDirection());
		if (rep[0] == 'X') {
			return Direction.dirValue(rep[1]);
		} else {
			return 0;
		}
	}

	public int getY() {
		Assert.check(!isSpecialDirection());
		if (rep[0] == 'Y') {
			return Direction.dirValue(rep[1]);
		} else {
			return 0;
		}
	}

	public int getZ() {
		Assert.check(!isSpecialDirection());
		if (rep[0] == 'Z') {
			return Direction.dirValue(rep[1]);
		} else {
			return 0;
		}
	}

	public boolean isPerpendicularDirection(Direction dir) {
		// This is a strange way to define perpendicular, but its what Selvin used...
		Assert.check(!isSpecialDirection());
		Assert.check(!dir.isSpecialDirection());

		return (rep[0] == dir.rep[0]);
	}

	public Direction getOpposite() {
		if (isSpecialDirection()) {
			return this;
		} else {
			if (rep[1] == '+') {
				return new Direction(rep[0], '-');
			} else {
				Assert.check(rep[1] == '-');
				return new Direction(rep[0], '+');
			}
		}
	}

	public String toString() {
		if (isSpecialDirection()) {
			if (this == UNDEFINED)
				return "<undefined>";
			if (this == ANY)
				return "<any>";
			if (this == ALL)
				return "<all>";
			if (this == OPPOSITE)
				return "<opposite>";
			if (this == SAME)
				return "<same>";
			if (this == PERPENDICULAR)
				return "<perpendicular>";
			Assert.badBranch("Bad direction: " + rep[0] + rep[1]);
		}

		return rep[0] + "" + rep[1];
	}

	public static DirectionSet allPossibleDirections(int dimensions) {
		DirectionSet dirs = new DirectionSet();

		for (int dir = 0; dir < numDirections(dimensions); dir++) {
			dirs.add(indexDirection(dir));
		}

		return dirs;
	}

	public Direction orient(Orientation orientation) {
		Position pos = canonicalize();
		Position oriented = orientation.rotate(pos);

		Direction dir = Direction.fromPosition(oriented);
		Report.message (3, 			"Orienting: "
				+ this
				+ ": "
				+ pos
				+ " ==> "
				+ oriented
				+ " == "
				+ dir);
		return dir;

	}

	public Direction minus(Direction rval) {
		Assert.badBranch ("not implemented");
		return null;
	}

	public Direction plus(Direction rval) {
		Assert.badBranch ("not implemented");
		return null;
	}
}
