package util;

import java.net.*;
import java.io.*;

public class StructuredReader extends java.io.BufferedReader {

	public StructuredReader(InputStream instream)
		throws MalformedURLException, IOException {
		super(new InputStreamReader(instream));
	}

	// effects: Advances to first character that is a stopper.
	public /*@non_null@*/
	String readUntilAny(/*@non_null@*/
	String stoppers) throws IOException {
		int ic;
		StringBuffer res = new StringBuffer();

		mark(1);

		while ((ic = read()) >= 0) {
			if (stoppers.indexOf(ic) != -1) {
				reset();
				return new String(res);
			} else {
				res.append((char) ic);
			}

			mark(1);
		}

		throw new EOFException(
			"Stop character not found before end of file: " + new String(res));
	} /*@nowarn NonNullResult@*/ /* ESC/Java reports spurious warnings for no return paths. */

	// effects: Advances to first characters after any stoppers.
	public /*@non_null@*/
	String readThroughAny(/*@non_null@*/
	String stoppers) throws IOException {
		int ic;
		StringBuffer res = new StringBuffer();

		while ((ic = read()) >= 0) {
			if (stoppers.indexOf(ic) != -1) {
				return new String(res);
			} else {
				res.append((char) ic);
			}
		}

		throw new EOFException(
			"Stop character not found before end of file: " + new String(res));
	} /*@nowarn NonNullResult@*/ /* ESC/Java reports spurious warnings for no return paths. */

	// Read the stream until a string matching match is found.  
	// VERY inefficient implementation...

	public /*@non_null@*/
	String readThroughString(/*@non_null@*/
	String match) throws IOException {
		int ic;
		StringBuffer res = new StringBuffer();

		while ((ic = read()) >= 0) {
			res.append((char) ic);

			if (res.length() >= match.length()) {
				int startIndex = res.length() - match.length();
				boolean matchOkay = true;

				for (int i = 0; i < match.length(); i++) {
					if (match.charAt(i) != res.charAt(startIndex + i)) {
						matchOkay = false;
						break;
					}
				}

				if (matchOkay) {
					return new String(res);
				}
			}
		}

		throw new EOFException(
			"Match string not found before end of file: " + new String(res));
	} /*@nowarn NonNullResult@*/ /* ESC/Java reports spurious warnings for no return paths. */

	private int readEitherInt(boolean allowCommas)
		throws IOException, NoNumberException {
		int ic;
		int readno = 0;
		boolean gotany = false;

		mark(1);

		while ((ic = read()) >= 0) {
			if (Character.isDigit((char) ic)) {
				readno = readno * 10 + (Character.getNumericValue((char) ic));
				gotany = true;
			} else if (allowCommas && ((char) ic == ',')) {
				; // ignore the commas 
			} else {
				reset();
				break;
			}

			mark(1);
		}

		if (gotany) {
			return readno;
		} else {
			throw new NoNumberException(ic);
		}
	}

	// Reads a double formates as [0-9]*.[E[+-][0-9]*]
	public double readDouble()
		throws IOException, NoNumberException, NumberFormatException {
		int ic;
		StringBuffer numstring = new StringBuffer();
		mark(1);

		while ((ic = read()) >= 0) {
			if (Character.isDigit((char) ic)
				|| (char) ic == 'E'
				|| (char) ic == '+'
				|| (char) ic == '-'
				|| (char) ic == '.') {
				numstring.append((char) ic);
			} else {
				reset();
				break;
			}

			mark(1);
		}

		if (numstring.length() == 0) {
			// System.err.println ("No number: " + numstring);			
			throw new NoNumberException (ic);
		} else {
			// System.err.println ("Read number: " + numstring);
			return Double.valueOf(new String(numstring)).doubleValue();
		}
	}

	public int readInt() throws IOException, NoNumberException {
		return readEitherInt(false);
	}

	public int readCommaInt() throws IOException, NoNumberException {
		return readEitherInt(true);
	}

	public int skipWhitespace() {
		int numread = 0;

		try {
			int ic;

			mark(1);

			while ((ic = read()) >= 0) {
				if (Character.isWhitespace((char) ic)) {
					numread++;
				} else {
					reset();
					break;
				}

				mark(1);
			}
		} catch (IOException e) {
			// No exception --- okay to end file skipping whitespace
			return numread;
		}

		return numread;
	}
}
